# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/06/2021"

import logging
import os
import shutil
import tempfile
import time
import pytest
from silx.gui import qt
from orangecontrib.tomwer.widgets.reconstruction.DarkRefAndCopyOW import (
    DarkRefAndCopyOW,
)

from tomwer.core.process.reconstruction.darkref.params import Method as DkrfMethod
from tomwer.gui.reconstruction.ftserie import FtserieWidget
from tomwer.gui.qtapplicationmanager import QApplicationManager
from tomwer.synctools.ftseries import QReconsParams
from silx.gui.utils.testutils import TestCaseQt
from tomwer.core.utils.scanutils import MockHDF5, MockEDF
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefs
from tomwer.core.process.reconstruction.darkref.settings import (
    DARKHST_PREFIX,
    REFHST_PREFIX,
)
import gc

_qapp = QApplicationManager()

logging.disable(logging.INFO)


class TestEDFDarkRefWidget(TestCaseQt):
    """class testing the DarkRefWidget"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._recons_params = QReconsParams()
        self.widget = DarkRefAndCopyOW(
            parent=None, reconsparams=self._recons_params.dkrf
        )

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        self.qapp.processEvents()
        gc.collect()
        TestCaseQt.tearDown(self)

    def testSyncRead(self):
        """Make sure any modification on the self._reconsParams is
        applied on the GUI"""
        rp = self._recons_params
        self.assertTrue(rp.dkrf["REFSRMV"] is False)
        self.assertFalse(
            self.widget.widget.mainWidget.tabGeneral._rmOptionCB.isChecked()
        )
        rp.dkrf["REFSRMV"] = True
        self.assertTrue(
            self.widget.widget.mainWidget.tabGeneral._rmOptionCB.isChecked()
        )

        pattern = self.widget.widget.mainWidget.tabExpert._refLE.text()
        newText = "popo.*"
        assert pattern != newText
        rp.dkrf["RFFILE"] = newText
        self.assertTrue(
            self.widget.widget.mainWidget.tabExpert._refLE.text() == newText
        )

    def testSyncWrite(self):
        """Test that if we edit through the :class:`DarkRefWidget` then the
        modification are fall back into the self._reconsParams"""
        rp = self._recons_params

        # test patterns
        pattern = self.widget.widget.mainWidget.tabExpert._refLE.text()
        newText = "popo.*"
        assert pattern != newText
        self.widget.widget.mainWidget.tabExpert._refLE.setText(newText)
        self.widget.widget.mainWidget.tabExpert._refLE.editingFinished.emit()
        qt.QApplication.instance().processEvents()
        self.assertTrue(rp.dkrf["RFFILE"] == newText)
        self.widget.widget.mainWidget.tabExpert._darkLE.setText(newText)
        self.widget.widget.mainWidget.tabExpert._darkLE.editingFinished.emit()
        qt.QApplication.instance().processEvents()
        self.assertTrue(rp.dkrf["DKFILE"] == newText)

        # test calc mode
        self.widget.widget.mainWidget.tabGeneral._darkWCB.setMode(DkrfMethod.none)
        self.widget.widget.mainWidget.tabGeneral._refWCB.setMode(DkrfMethod.median)
        self.assertTrue(rp.dkrf["DARKCAL"] == DkrfMethod.none)
        self.assertTrue(rp.dkrf["REFSCAL"] == DkrfMethod.median)

        # test options
        cuRm = self.widget.widget.mainWidget.tabGeneral._rmOptionCB.isChecked()
        self.widget.widget.mainWidget.tabGeneral._rmOptionCB.setChecked(not cuRm)
        self.assertTrue(rp.dkrf["REFSRMV"] == (not cuRm))
        self.assertTrue(rp.dkrf["DARKRMV"] == (not cuRm))

        cuSkip = self.widget.widget.mainWidget.tabGeneral._skipOptionCB.isChecked()
        self.widget.widget.mainWidget.tabGeneral._skipOptionCB.setChecked(not cuSkip)
        # warning : here value of skip and overwrite are of course inverse
        self.assertTrue(rp.dkrf["DARKOVE"] == cuSkip)
        self.assertTrue(rp.dkrf["REFSOVE"] == cuSkip)

    def testBehaviorWithFtserie(self):
        """
        Make sure modification on ftserie 'Dark and flat field' tab are
        correctly take into account in this case they are working on the same
        QReconsParams object
        """
        ftserie = FtserieWidget(recons_params=self._recons_params)

        cuRm = ftserie.getReconsParamSetEditor()._dkRefWidget._qcbRmRef.isChecked()
        cuSkip = ftserie.getReconsParamSetEditor()._dkRefWidget._qcbSkipRef.isChecked()
        calcDK = ftserie.getReconsParamSetEditor()._dkRefWidget._qcbDKMode.getMode()
        calcRef = ftserie.getReconsParamSetEditor()._dkRefWidget._qcbRefMode.getMode()
        patternRef = (
            ftserie.getReconsParamSetEditor()._dkRefWidget._qleRefsPattern.text()
        )
        patternDK = ftserie.getReconsParamSetEditor()._dkRefWidget._qleDKPattern.text()

        # make sure initial status are the same between ftserie and darkref
        dkrfOWMainWidget = self.widget.widget.mainWidget
        self.assertTrue(dkrfOWMainWidget.tabGeneral._rmOptionCB.isChecked() == cuRm)
        self.assertTrue(dkrfOWMainWidget.tabGeneral._skipOptionCB.isChecked() == cuSkip)
        self.assertTrue(dkrfOWMainWidget.tabGeneral._darkWCB.getMode() == calcDK)
        self.assertTrue(dkrfOWMainWidget.tabGeneral._refWCB.getMode() == calcRef)
        self.assertTrue(dkrfOWMainWidget.tabExpert._refLE.text() == patternRef)
        self.assertTrue(dkrfOWMainWidget.tabExpert._darkLE.text() == patternDK)

        # change parameters
        ftserie.getReconsParamSetEditor()._dkRefWidget._qcbRmRef.setChecked(not cuRm)
        ftserie.getReconsParamSetEditor()._dkRefWidget._qcbSkipRef.setChecked(
            not cuSkip
        )
        ftserie.getReconsParamSetEditor()._dkRefWidget._qcbDKMode.setMode("none")
        ftserie.getReconsParamSetEditor()._dkRefWidget._qcbRefMode.setMode("average")
        text_ref_pattern = "toto.*"
        text_dk_pattern = "[maestro]?."
        ftserie.getReconsParamSetEditor()._dkRefWidget._qleDKPattern.setText(
            text_dk_pattern
        )
        ftserie.getReconsParamSetEditor()._dkRefWidget._qleDKPattern.editingFinished.emit()
        qt.QApplication.instance().processEvents()
        ftserie.getReconsParamSetEditor()._dkRefWidget._qleRefsPattern.setText(
            text_ref_pattern
        )
        ftserie.getReconsParamSetEditor()._dkRefWidget._qleRefsPattern.editingFinished.emit()
        qt.QApplication.instance().processEvents()

        # check modification are well take into account
        self.assertTrue(
            dkrfOWMainWidget.tabGeneral._rmOptionCB.isChecked() == (not cuRm)
        )
        self.assertTrue(self._recons_params.dkrf["DARKRMV"] == (not cuRm))
        self.assertTrue(
            dkrfOWMainWidget.tabGeneral._skipOptionCB.isChecked() == (not cuSkip)
        )
        self.assertTrue(self._recons_params.dkrf["DARKOVE"] == cuSkip)
        self.assertTrue(
            dkrfOWMainWidget.tabGeneral._darkWCB.getMode() == DkrfMethod.none
        )
        self.assertTrue(self._recons_params.dkrf["DARKCAL"] == DkrfMethod.none)
        self.assertTrue(
            dkrfOWMainWidget.tabGeneral._refWCB.getMode() == DkrfMethod.average
        )
        self.assertTrue(self._recons_params.dkrf["REFSCAL"] == DkrfMethod.average)
        self.assertTrue(self._recons_params.dkrf["DKFILE"] == text_dk_pattern)
        self.assertTrue(self._recons_params.dkrf["RFFILE"] == text_ref_pattern)
        self.assertTrue(dkrfOWMainWidget.tabExpert._refLE.text() == text_ref_pattern)
        self.assertTrue(dkrfOWMainWidget.tabExpert._darkLE.text() == text_dk_pattern)


@pytest.mark.skip("Fail on CI")
class TestDarkRefCopyWithEDFAndHDF5(TestCaseQt):
    """Test the DarkRefCopy orange widget behaviour"""

    def setUp(self) -> None:
        TestCaseQt.setUp(self)
        self._folder = tempfile.mkdtemp()

        # define scans to be treated
        hdf5_mock_with_refs = MockHDF5(
            scan_path=os.path.join(self._folder, "h5_with_refs"),
            n_proj=10,
            n_ini_proj=10,
            dim=20,
            create_ini_ref=True,
            create_ini_dark=True,
            create_final_ref=False,
        )
        self.hdf5_acquisition_with_refs = hdf5_mock_with_refs.scan

        hdf5_mock_without_refs = MockHDF5(
            scan_path=os.path.join(self._folder, "h5_without_refs"),
            n_proj=10,
            n_ini_proj=10,
            dim=20,
            create_ini_ref=False,
            create_ini_dark=False,
        )
        self.hdf5_acquisition_without_refs = hdf5_mock_without_refs.scan

        hdf5_mock_without_refs_incoherent_dim = MockHDF5(
            scan_path=os.path.join(self._folder, "h5_without_refs_different_dim"),
            n_proj=10,
            n_ini_proj=10,
            dim=21,
            create_ini_ref=False,
            create_ini_dark=False,
        )
        self.hdf5_acquisition_without_refs_incoherent_dim = (
            hdf5_mock_without_refs_incoherent_dim.scan
        )

        edf_mock_without_ref = MockEDF(
            scan_path=os.path.join(self._folder, "edf_without_refs"), dim=20, n_radio=20
        )
        self.edf_acquisition_without_ref = EDFTomoScan(
            scan=edf_mock_without_ref.scan_path
        )

        # processes set up
        self._recons_params = QReconsParams()
        self.widget = DarkRefAndCopyOW(
            parent=None, reconsparams=self._recons_params.dkrf
        )
        self.widget.setForceSync(True)
        self.widget.show()

    def tearDown(self) -> None:
        time.sleep(0.5)
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self._folder)
        TestCaseQt.tearDown(self)

    def testCopyInactive(self):
        self.widget.setCopyActive(False)

        self.widget.process(self.hdf5_acquisition_with_refs)
        self.assertTrue(os.path.exists(self.hdf5_acquisition_with_refs.process_file))

        # make sure dark has been processed for the one with ref
        self.assertEqual(
            len(
                DarkRefs.get_flats_frm_process_file(
                    self.hdf5_acquisition_with_refs.process_file
                )
            ),
            1,
        )
        self.assertEqual(
            len(
                DarkRefs.get_darks_frm_process_file(
                    self.hdf5_acquisition_with_refs.process_file
                )
            ),
            1,
        )
        # make sure no dark or flats exists for the hdf5 without refs
        for scan in (
            self.hdf5_acquisition_without_refs_incoherent_dim,
            self.hdf5_acquisition_without_refs,
        ):
            self.widget.process(scan)
            self.assertFalse(os.path.exists(scan.process_file))

        # make sure no dark or flats exists for the edf one
        self.widget.process(self.edf_acquisition_without_ref)

    def testCopyActive(self):
        self.widget.setCopyActive(True)
        self.widget.setModeAuto(True)

        self.widget.process(self.hdf5_acquisition_with_refs)
        self.assertTrue(os.path.exists(self.hdf5_acquisition_with_refs.process_file))
        # 1. make sure dark has been processed for the one with ref
        self.assertEqual(
            len(
                DarkRefs.get_flats_frm_process_file(
                    self.hdf5_acquisition_with_refs.process_file
                )
            ),
            1,
        )
        self.assertEqual(
            len(
                DarkRefs.get_darks_frm_process_file(
                    self.hdf5_acquisition_with_refs.process_file
                )
            ),
            1,
        )
        self.assertTrue(self.widget.hasDarkStored())
        self.assertTrue(self.widget.hasFlatStored())

        # 2. make sure copy has been processed for the the 'compatible hdf5'
        self.widget.process(self.hdf5_acquisition_without_refs)
        self.assertTrue(os.path.exists(self.hdf5_acquisition_without_refs.process_file))
        self.assertEqual(
            len(
                DarkRefs.get_flats_frm_process_file(
                    self.hdf5_acquisition_without_refs.process_file
                )
            ),
            2,
        )
        self.assertEqual(
            len(
                DarkRefs.get_darks_frm_process_file(
                    self.hdf5_acquisition_without_refs.process_file
                )
            ),
            1,
        )

        # 3. make sure copy has been processed for the the 'compatible edf'
        self.widget.process(self.edf_acquisition_without_ref)
        self.assertEqual(
            len(
                DarkRefs.getRefHSTFiles(
                    self.edf_acquisition_without_ref.path, prefix=REFHST_PREFIX
                )
            ),
            2,
        )
        self.assertEqual(
            len(
                DarkRefs.getDarkHSTFiles(
                    self.edf_acquisition_without_ref.path, prefix=DARKHST_PREFIX
                )
            ),
            1,
        )

        # 4. make sure process but no copy made if incompatible size
        self.widget.process(self.hdf5_acquisition_without_refs_incoherent_dim)
        self.assertFalse(
            os.path.exists(
                self.hdf5_acquisition_without_refs_incoherent_dim.process_file
            )
        )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "25/05/2018"


from silx.gui import qt
from tomwer.gui.qfolderdialog import QScanDialog
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.control.datalist import GenericDataList
import logging

logger = logging.getLogger(__name__)


class ScanSelectorWidget(qt.QWidget):
    """Widget used to select a scan on a list"""

    sigSelectionChanged = qt.Signal(list)
    """Signal emitted when the selection changed"""

    sigUpdated = qt.Signal()
    """signal emitted when the scan list change"""

    def __init__(self, parent=None):
        def getAddAndRmButtons():
            lLayout = qt.QHBoxLayout()
            w = qt.QWidget(self)
            w.setLayout(lLayout)
            self._addButton = qt.QPushButton("Add")
            self._addButton.clicked.connect(self._callbackAddFolder)
            self._rmButton = qt.QPushButton("Remove")
            self._rmButton.clicked.connect(self._callbackRemoveFolder)

            spacer = qt.QWidget(self)
            spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
            lLayout.addWidget(spacer)
            lLayout.addWidget(self._addButton)
            lLayout.addWidget(self._rmButton)

            return w

        def getSendButton():
            lLayout = qt.QHBoxLayout()
            widget = qt.QWidget(self)
            widget.setLayout(lLayout)
            self._sendButton = qt.QPushButton("Select")
            self._sendButton.clicked.connect(self._selectActiveScan)

            spacer = qt.QWidget(self)
            spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
            lLayout.addWidget(spacer)
            lLayout.addWidget(self._sendButton)

            return widget

        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self.dataList = GenericDataList(parent=self)
        self.dataList.setSelectionMode(qt.QAbstractItemView.ExtendedSelection)
        self.layout().addWidget(self.dataList)
        self.layout().addWidget(getAddAndRmButtons())
        self.layout().addWidget(getSendButton())
        self.setAcceptDrops(True)

    def add(self, scan):
        self.dataList.add(scan)
        self.sigUpdated.emit()

    def remove(self, scan):
        self.dataList.remove(scan)
        self.sigUpdated.emit()

    def n_scan(self) -> int:
        return len(self.dataList.items)

    def selectAll(self):
        self.dataList.selectAll()

    def _callbackAddFolder(self):
        """ """
        dialog = QScanDialog(self, multiSelection=True)

        if not dialog.exec_():
            dialog.close()
            return

        for folder in dialog.files_selected():
            self.add(folder)
        self.sigUpdated.emit()

    def _selectActiveScan(self):
        sItem = self.dataList.selectedItems()
        if sItem and len(sItem) >= 1:
            selection = [_item.text() for _item in sItem]
            self.sigSelectionChanged.emit(list(selection))
        else:
            logger.warning("No active scan detected")

    def _callbackRemoveFolder(self):
        """ """
        selectedItems = self.dataList.selectedItems()
        if selectedItems is not None:
            for item in selectedItems:
                self.dataList.remove_item(item)
        self.sigUpdated.emit()

    def setActiveScan(self, scan):
        """
        set the given scan as the active one

        :param scan: the scan to set active
        :type scan: Union[str, TomoBase]
        """
        scanID = scan
        if isinstance(scan, TomwerScanBase):
            scanID = scan.path
        self.dataList.setCurrentItem(self.dataList.items[scanID])



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2018"


from silx.gui import qt
import logging
from tomoscan.scanbase import _FOV
from tomwer.synctools.ftseries import QReconsParams
from tomwer.gui.reconstruction.ftserie.h5editor.h5structseditor import H5StructsEditor

logger = logging.getLogger(__name__)


class FTAxisWidget(H5StructsEditor, qt.QWidget):
    """
    Widget containing all the information to edit the AXIS parameters

    :param reconsparams: reconstruction parameters edited by the widget
    """

    def __init__(self, reconsparams, parent=None):
        qt.QWidget.__init__(self, parent)
        H5StructsEditor.__init__(self, structsManaged=("FT", "FTAXIS"))

        assert isinstance(reconsparams, QReconsParams)

        self._recons_params = None
        self.setReconsParams(recons_params=reconsparams)

        # build gui
        self.setLayout(qt.QVBoxLayout())

        self._qcbHalfAcq = qt.QCheckBox("Half acquisition (HA)", parent=self)
        self.layout().addWidget(self._qcbHalfAcq)

        self._doAxisCorrection = qt.QCheckBox("apply axis correction", parent=self)
        self._doAxisCorrection.setToolTip(
            "If activated then ask to pyhst to " "do axis translation correction"
        )
        self.layout().addWidget(self._doAxisCorrection)
        # for now the axis correction (one value per projection - when detector
        # moves is not managed)
        self._doAxisCorrection.hide()

        self._useTomwerAxis = qt.QCheckBox(
            "use center of rotation computed upstream (if any)", parent=self
        )
        self._useTomwerAxis.setToolTip(
            "If selected, the value computed "
            "upstream in this flow will be used. \n"
            "This option has priority over "
            "other options"
        )
        self.layout().addWidget(self._useTomwerAxis)

        self._useOldTomwerAxis = qt.QCheckBox(
            "use center of rotation calculation from previous session (if any)",
            parent=self,
        )
        self._useOldTomwerAxis.setToolTip(
            "If the scan has already been processed by tomwer, "
            "the value of the center of rotation used had been recorded. \n"
        )
        self.layout().addWidget(self._useOldTomwerAxis)

        # add spacer
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        # link the check box
        self.linkGroupWithH5Variable(
            group=self._doAxisCorrection,
            structID="FTAXIS",
            h5ParamName="DO_AXIS_CORRECTION",
            setter=self._setDoAxisCorrection,
            getter=self._getDoAxisCorrection,
        )
        self.linkGroupWithH5Variable(
            group=self._useTomwerAxis,
            structID="FTAXIS",
            h5ParamName="USE_TOMWER_AXIS",
            setter=self._setUseTomwerAxis,
            getter=self._getUseTomwerAxis,
        )
        self.linkGroupWithH5Variable(
            group=self._useOldTomwerAxis,
            structID="FTAXIS",
            h5ParamName="TRY_USE_OLD_TOMWER_AXIS",
            setter=self._setTryUseOldTomwerAxis,
            getter=self._getTryUseOldTomwerAxis,
        )

        self.linkCheckboxWithH5Variable(
            self._qcbHalfAcq, structID="FT", h5ParamName="HALF_ACQ"
        )

        # signal / Slot connection
        self._doAxisCorrection.toggled.connect(self._doAxisCorrectionChanged)
        self._useTomwerAxis.toggled.connect(self._useTomwerAxisChanged)
        self._useOldTomwerAxis.toggled.connect(self._useOldTomwerAxisChanged)
        self._qcbHalfAcq.toggled.connect(self._HALFACQChanged)

    def setReconsParams(self, recons_params):
        """
        Define the `.ReconsParams` that the widget edit. Those parameters will
        be copied in all the TomoBase process by the widget

        :param recons_params: tomographic parameter to edit
        :type recons_params: `.ReconsParams`
        """
        assert isinstance(recons_params, QReconsParams)

        if self._recons_params:
            self._recons_params.sigChanged.disconnect(self._update_params)
        self._recons_params = recons_params
        self.load(self._recons_params)
        self._recons_params.sigChanged.connect(self._update_params)

    def _update_params(self):
        """Update all parameter"""
        self.load(self._recons_params)

    def _setDoAxisCorrection(self, doAxisCorrrection: bool) -> None:
        assert type(doAxisCorrrection) is bool
        self._doAxisCorrection.setChecked(doAxisCorrrection)

    def _getDoAxisCorrection(self) -> bool:
        return self._doAxisCorrection.isChecked()

    def _setUseTomwerAxis(self, useTomwerAxis: bool) -> None:
        assert type(useTomwerAxis) is bool
        self._useTomwerAxis.setChecked(useTomwerAxis)

    def _getUseTomwerAxis(self) -> bool:
        return self._useTomwerAxis.isChecked()

    def _setTryUseOldTomwerAxis(self, useOldTomwerAxis: bool) -> None:
        assert type(useOldTomwerAxis) is bool
        self._useOldTomwerAxis.setChecked(useOldTomwerAxis)

    def _getTryUseOldTomwerAxis(self) -> bool:
        return self._useOldTomwerAxis

    def _doAxisCorrectionChanged(self, *args, **kwargs) -> None:
        doAxisCorrection = self._getDoAxisCorrection()
        self._recons_params["FTAXIS"]["DO_AXIS_CORRECTION"] = doAxisCorrection

    def _useTomwerAxisChanged(self, *args, **kwargs) -> None:
        self._recons_params["FTAXIS"]["USE_TOMWER_AXIS"] = self._getUseTomwerAxis()

    def _useOldTomwerAxisChanged(self, *args, **kwargs) -> None:
        self._recons_params["FTAXIS"][
            "TRY_USE_OLD_TOMWER_AXIS"
        ] = self._getTryUseOldTomwerAxis()

    def _HALFACQChanged(self, b):
        if int(b) != self._recons_params["FT"]["HALF_ACQ"]:
            self._recons_params["FT"]["HALF_ACQ"] = int(b)

    def setScan(self, scan):
        fov = scan.field_of_view
        if fov is not None:
            fov = _FOV.from_value(fov)
            self._qcbHalfAcq.setChecked(fov == _FOV.HALF)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""This module provides global definitions to prepare fasttomo input as HDF5 file.
of reconstruction parameters structure and should be a mirror of the octave one:
/sware/pub/octave-tomwer/suse82/m/tomotools/DEV/defineGLOBALS.m
"""

__authors__ = ["C. Nemoz", "H.Payno"]
__license__ = "MIT"
__date__ = "25/05/2016"

import os
import logging
from tomwer.core.process.reconstruction.ftseries.params import ReconsParams
import copy

logger = logging.getLogger(__name__)

try:
    from silx.io.octaveh5 import Octaveh5
except ImportError as e:
    logger.error("Module " + __name__ + " requires silx.io.octaveh5")
    raise e


class H5MissingParameters(Exception):
    """Exception launch when some h5 parameters are missing from a structure"""

    def __init__(self, missingStruc, missingVar):
        super(H5MissingParameters, self).__init__()
        self.missingStructures = missingStruc
        self.missingVariables = missingVar

    def __str__(self):
        mess = "File is missing some structures and/or variables."
        mess += "Values of those has been setted to default values. "
        mess += "Please make sure they are correct. Missing structures are "
        mess += " missing variables \n: %s" % self.missingVariables
        mess += " missing structures \n: %s" % self.missingStructures
        return mess


class FastSetupAll(object):
    OFFV = "pyhst2"
    h5_prefix = "ftinput_"  # prefix of h5 files to consider it is an octave input

    DEFAULT_VALUES = ReconsParams()

    def __init__(self):
        self.tmph5 = "octave_input.h5"
        self.structures = ReconsParams()

    @staticmethod
    def getDefaultValues(structID):
        """Return a deep copy of the struct ID param name / value dict"""
        if structID == "FT":
            return copy.deepcopy(FastSetupAll.DEFAULT_VALUES.ft.to_dict())
        elif structID == "FTAXIS":
            return copy.deepcopy(FastSetupAll.DEFAULT_VALUES.axis.to_dict())
        elif structID == "PAGANIN":
            return copy.deepcopy(FastSetupAll.DEFAULT_VALUES.paganin.to_dict())
        elif structID == "PYHSTEXE":
            return copy.deepcopy(FastSetupAll.DEFAULT_VALUES.pyhst.to_dict())
        elif structID == "BEAMGEO":
            return copy.deepcopy(FastSetupAll.DEFAULT_VALUES.beam_geo.to_dict())
        elif structID == "DKRF":
            return copy.deepcopy(FastSetupAll.DEFAULT_VALUES.dkrf.to_dict())
        else:
            raise ValueError("%s has no default values" % structID)

    def readAll(self, filn, targetted_octave_version):
        if not os.path.isfile(filn):
            raise IOError("given path is not a file %s" % filn)
        reader = Octaveh5(targetted_octave_version).open(filn)
        self.resetDefaultStructures()
        # get minimal groups (existing groups and default groups)
        for st in list(reader.file):
            if st in self.structures.managed_params:
                self.structures[st].load_from_dict(reader.get(st))
            else:
                if st in self.structures.unmanaged_params:
                    self.structures._remove_unmanaged_param(param=st)

                self.structures._add_unmanaged_param(param=st, value=reader.get(st))
            assert st in self.structures.all_params

    def resetDefaultStructures(self):
        self.structures.load_from_dict(ReconsParams().to_dict())
        self.structures._reset_unmanaged_params()

    def writeAll(self, filn, targetted_octave_version):
        if os.path.isfile(filn) is True:
            os.remove(filn)

        writer = Octaveh5(targetted_octave_version)
        writer.open(filn, "a")
        structures = self.structures.to_dict()
        for s in structures:
            assert type(structures[s]) is dict
            writer.write(s, structures[s])
        writer.close()

    @staticmethod
    def getAllDefaultStructures():
        return ReconsParams()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys
from silx.gui import qt
import argparse
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.reconstruction.nabu.slices import NabuDialog
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.process.reconstruction.nabu.nabuslices import run_slices_reconstruction
from tomwer.gui.visualization.dataviewer import DataViewer as _DataViewer
from tomwer.core.utils.resource import increase_max_number_file
import signal

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


def getinputinfo():
    return "tomwer nabu [scan_path]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


class DataViewer(_DataViewer):
    """The data viewer with an update when F5 is press"""

    closed = qt.Signal()
    """signal emitted when the widget is closing"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setAttribute(qt.Qt.WA_DeleteOnClose)

    def keyPressEvent(self, event):
        if event.key() is qt.Qt.Key_F5:
            self._updateDisplay()

    def event(self, event) -> bool:
        if event.type() == qt.QEvent.DeferredDelete:
            self.cleanBeforeQuit()
            self.closed.emit()
        return super().event(event)


class NabuProcessingThread(qt.QThread):
    def init(self, scan, config, dry_run=False):
        self._scan = scan
        self._config = config
        self._dry_run = dry_run

    def run(self):
        run_slices_reconstruction(
            scan=self._scan, config=self._config, dry_run=self._dry_run
        )


class _CORWidget(qt.QWidget):
    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(qt.QLabel("center of rotation", self))
        self._corCN = qt.QCheckBox("auto", self)
        self.layout().addWidget(self._corCN)
        self._corQLE = qt.QLineEdit("0", self)
        self._corQLE.setValidator(qt.QDoubleValidator(self))
        self.layout().addWidget(self._corQLE)
        self._corQLE.setToolTip("Value should be in -width/2, width/2")

        # set up
        self._corCN.setChecked(True)
        self._corQLE.setEnabled(False)

        # connect signal / slot
        self._corCN.toggled.connect(self._corQLE.setDisabled)

    def getCOR(self):
        if self._corCN.isChecked():
            return "auto"
        else:
            return float(self._corQLE.text())


class NabuWindow(NabuDialog):
    def __init__(self, parent=None):
        NabuDialog.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        self.setAttribute(qt.Qt.WA_DeleteOnClose)

        self.__scan = None
        self.__processingThread = NabuProcessingThread()
        self._dry_run = False

        # add center of rotation widget
        self._corWidget = _CORWidget(parent=self)
        rw = self._widget._mainWidget._configuration._reconstructionWidget
        rw.layout().addWidget(self._corWidget, 99, 0, 1, 2)

        self._dataViewer = None
        self.setModal(False)

        # add view button
        self._showPB = qt.QPushButton("show scan", self)
        self._buttons.addButton(self._showPB, qt.QDialogButtonBox.ActionRole)

        # connect signal / slot
        self._computePB.released.connect(self._requireComputation)
        self._showPB.released.connect(self._showScan)

        # export API
        self.getCOR = self._corWidget.getCOR

    def _showScan(self):
        self._getDataViewer().setScan(self.__scan)
        self._getDataViewer().show()
        self._getDataViewer().raise_()

    def _getDataViewer(self):
        if self._dataViewer is None:
            self._dataViewer = DataViewer(parent=None)
            self._dataViewer.closed.connect(self._viewerDeleted)

        return self._dataViewer

    def _viewerDeleted(self):
        self._dataViewer = None

    def setScan(self, scan):
        self.__scan = scan
        NabuDialog.setScan(self, scan)

    def getScan(self):
        return self.__scan

    def getConfiguration(self):
        config = super().getConfiguration()
        cor = self.getCOR()
        if cor != "auto":
            # move the cor value to the nabu reference
            cor = cor + self.__scan.dim_1 // 2
        config["reconstruction"]["rotation_axis_position"] = str(cor)
        return config

    def _requireComputation(self, *arg, **kwargs):
        _logger.info("start computation")
        scan = self.getScan()
        if scan is not None:
            qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
            self._launchComputation(scan=scan, config=self.getConfiguration())

    def _computationEnded(self):
        _logger.info("end computation")
        qt.QApplication.restoreOverrideCursor()
        if self._dataViewer is not None:
            self._dataViewer.setScan(self.__scan)

    def getProcessingThread(self):
        return self.__processingThread

    def _launchComputation(self, scan, config):
        # update the processing thread
        thread = self.getProcessingThread()
        thread.init(scan=scan, config=config, dry_run=self._dry_run)
        thread.finished.connect(self._computationEnded)

        # start processing
        thread.start(priority=qt.QThread.LowPriority)

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    def close(self):
        super().close()
        if self._dataViewer is not None:
            self._dataViewer.close()
            self._dataViewer = None

    def _aboutToQuit(self):
        if self._dataViewer is not None:
            self._dataViewer.cleanBeforeQuit()


def main(argv):
    import os

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path", help="Data file to show (h5 file, edf files, spec files)"
    )
    parser.add_argument(
        "--entry", default=None, help="An entry should be specify for hdf5 files"
    )
    parser.add_argument(
        "--output-dir", help="output directory of the reconstruction(s)"
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    parser.add_argument(
        "--dry-run",
        dest="dry_run",
        action="store_true",
        default=False,
        help="Only create the .nabu file and print the configuration used",
    )
    options = parser.parse_args(argv[1:])

    increase_max_number_file()

    app = qt.QApplication.instance() or qt.QApplication([])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler
    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()
    options.scan_path = options.scan_path.rstrip(os.path.sep)
    scan = ScanFactory.create_scan_object(options.scan_path, entry=options.entry)
    widget = NabuWindow(parent=None)
    widget.setScan(scan)
    scan.set_process_index_frm_tomwer_process_file()
    if options.output_dir is not None:
        widget.setOutputDir(options.output_dir)
    # for the application we run for the reconstruction to be finished
    # to give back hand to the user
    widget.set_dry_run(options.dry_run)
    widget.setWindowTitle("Nabu reconstruction")
    widget.setWindowIcon(icons.getQIcon("tomwer"))
    splash.finish(widget)
    widget.show()
    qt.QApplication.instance().aboutToQuit.connect(widget._aboutToQuit)

    app.exec_()


if __name__ == "__main__":
    main(sys.argv)



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "08/06/2017"


import os
from queue import Queue
from silx.gui import qt
from tomwer.core import settings
from tomwer.core import utils
from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.process.reconstruction.ftseries.params import fastsetupdefineglobals
from tomwer.core.process.reconstruction.ftseries import (
    run_reconstruction_from_recons_params,
)
from tomwer.core.utils import logconfig
from tomwer.core.utils.Singleton import singleton
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.web.client import OWClient
from processview.core.superviseprocess import SuperviseProcess
from processview.core.manager import ProcessManager, DatasetState
import logging

logger = logging.getLogger(__name__)


@singleton
class ReconstructionStack(qt.QObject, Queue, OWClient, SuperviseProcess):
    """
    Manage a stack of ftseries reconstruction
    """

    sigReconsFinished = qt.Signal(TomwerScanBase)
    sigReconsFailed = qt.Signal(TomwerScanBase)
    sigReconsMissParams = qt.Signal(TomwerScanBase)
    sigReconsStarted = qt.Signal(TomwerScanBase)

    def __init__(self, process_id=None):
        qt.QObject.__init__(self)
        Queue.__init__(self)
        OWClient.__init__(self)
        SuperviseProcess.__init__(self, process_id=process_id)
        self.reconsThread = _ReconsFtSeriesThread(process_id=process_id)
        self.reconsThread.sigThReconsFinished.connect(self._dealWithFinishedRecons)
        self.reconsThread.sigThReconsFailed.connect(self._dealWithFailedRecons)
        self.reconsThread.sigThMissingParams.connect(self._dealWithThMissingParams)
        self._forceSync = False

    def add(self, scan, inputs):
        """
        add a reconstruction and will run it as soon as possible

        :param list slices: the list of slices to reconstruct
        :param scan: the folder of the acquisition to reconstruct
        :type scan: :class:`.TomoBase`
        :param dict inputs: parameters of the reconstruction
        """
        ProcessManager().notify_dataset_state(
            dataset=scan,
            process=self,
            state=DatasetState.PENDING,
        )
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(f"{scan} is expected to be an instance of {TomwerScanBase}")
        # TODO: reconsparam could be taken dircetly from scan
        Queue.put(self, (scan, inputs))
        if self.canExecNext():
            self.execNext()

    def execNext(self):
        """Launch the next reconstruction if any"""
        if Queue.empty(self):
            return

        assert not self.reconsThread.isRunning()
        # TODO: reconsparam could be taken dircetly from scan
        scan, inputs = Queue.get(self)
        self.sigReconsStarted.emit(scan)
        self.reconsThread.init(scan, inputs)
        self.reconsThread.start()
        if self._forceSync is True:
            self.reconsThread.wait()

    def canExecNext(self):
        """
        Can we launch an ftserie reconstruction.
        Reconstruction can't be runned in parallel

        :return: True if no reconstruction is actually running
        """
        return not self.reconsThread.isRunning()

    def _dealWithFinishedRecons(self, scan):
        assert isinstance(scan, TomwerScanBase)
        info = "reconstruction %s is finished" % scan.path
        logger.info(info)
        self.sigReconsFinished.emit(scan)
        self.execNext()

    def _dealWithThMissingParams(self, scan):
        self.sigReconsMissParams.emit(scan)
        self.execNext()

    def _dealWithFailedRecons(self, scan):
        self.sigReconsFailed.emit(scan)
        self.execNext()

    def setMockMode(self, b):
        self.reconsThread.setMockMode(b)
        self.execNext()

    def setForceSync(self, b=True):
        self._forceSync = b

    def stop(self):
        """stop the stack and get back the thread"""
        self.queue.clear()
        self.blockSignals(True)
        self.reconsThread.blockSignals(True)
        self.reconsThread.wait()


class _ReconsFtSeriesThread(OWClient, qt.QThread, SuperviseProcess):
    """
    Simple thread launching ftseries reconstructions
    """

    copyH5FileReconsIntoFolder = False
    """if True then copy the file used for reconstruction into the scan
    folder under the copyH5ReconsName"""

    copyH5ReconsName = "octave_FT_params.h5"
    """Name under wich the reconstruction parameters will be saved if
    'copyH5FileReconsIntoFolder' is active"""

    sigThReconsFinished = qt.Signal(TomwerScanBase)
    "Emitted if reconstruction ended with success"
    sigThReconsFailed = qt.Signal(TomwerScanBase)
    "Emitted if reconstruction failed"
    sigThMissingParams = qt.Signal(TomwerScanBase)
    "Emitted if missing some reconstruction parameters"

    _mockMode = False

    def __init__(self, process_id=None):
        OWClient.__init__(self)
        # handle Qt different constructor requests
        try:
            qt.QThread.__init__(self, process_id=process_id)
        except TypeError:
            qt.QThread.__init__(self)
        SuperviseProcess.__init__(self, process_id=process_id)
        self.scan = None
        self.inputs = None

    def run(self):
        if self.scan is None or self.inputs is None:
            mess = "reconstruction not initialized. Can' reconstruct"
            logger.warning(mess)
            ProcessManager().notify_dataset_state(
                dataset=self.scan, process=self, state=DatasetState.FAILED, details=mess
            )

            self.sigThReconsFailed.emit(mess)
            return

        if not os.path.isdir(self.scan.path):
            mess = "%s is not a valid fodler, can't reconstruct"
            ProcessManager().notify_dataset_state(
                dataset=self.scan, process=self, state=DatasetState.FAILED, details=mess
            )
            self.sigThReconsFailed.emit(mess)
            return

        self._processRecons()

    def init(self, scan, inputs):
        assert isinstance(scan, TomwerScanBase)
        self.scan = scan
        self.inputs = inputs

    def _processRecons(self):
        if not isinstance(self.inputs, dict):
            raise TypeError(f"{self.inputs} is expected to be a dict")
        ProcessManager().notify_dataset_state(
            dataset=self.scan,
            process=self,
            state=DatasetState.ON_GOING,
        )
        self._noticeStart()
        # check is on low memory in lbsram
        if (
            settings.isOnLbsram(self.scan)
            and utils.isLowOnMemory(settings.get_lbsram_path()) is True
        ):
            # if computer is running into low memory in lbsram skip reconstruction
            mess = "low memory, skip reconstruction for " + self.scan.path
            logger.processSkipped(mess)
            ProcessManager().notify_dataset_state(
                dataset=self.scan,
                process=self,
                state=DatasetState.SKIPPED,
                details=mess,
            )

            self.sigThReconsFinished.emit(self.scan)
        else:
            try:
                # deal with mock reconstruction
                if self._mockMode is True:
                    MockEDF.mockReconstruction(self.scan.path)
                else:
                    if not ("pyhst2_params" in self.inputs):
                        raise KeyError("'pyhst2_params' key is expected")
                    if not isinstance(self.inputs["pyhst2_params"], dict):
                        raise TypeError(
                            "value associated to pyhst2_params is expected to be a dict"
                        )
                    run_reconstruction_from_recons_params(
                        scan=self.scan, inputs=self.inputs
                    )
            except fastsetupdefineglobals.H5MissingParameters:
                ProcessManager().notify_dataset_state(
                    dataset=self.scan,
                    process=self,
                    state=DatasetState.FAILED,
                )
                self.sigThMissingParams.send(self, self.scan)
            else:
                ProcessManager().notify_dataset_state(
                    dataset=self.scan,
                    process=self,
                    state=DatasetState.SUCCEED,
                )

        self.sigThReconsFinished.emit(self.scan)

    def _noticeStart(self):
        info = "start reconstruction of %s" % self.scan.path
        logger.info(info, extra={logconfig.SCAN_ID: self.scan.path})

    def setMockMode(self, b):
        """If the mock mode is activated then during reconstruction won't call
        Octave script for reconstruction but will generate some output files
        according to convention

        :param boolean b: True if we want to active the mock mode
        """
        self._mockMode = b



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "26/10/2021"


import functools
import os
import copy
import logging
from processview.core.manager.manager import ProcessManager

from tomwer.core.scan.futurescan import FutureTomwerScan
from tomwer.core.utils.slurm import is_slurm_available

try:
    from nabu.pipeline.fullfield.local_reconstruction import (  # noqa F401
        ChunkedReconstructor,
    )
except ImportError as e:
    has_nabu = False
    _err_import_nabu = e
else:
    has_nabu = True
    _err_import_nabu = None
from silx.utils.enum import Enum as _Enum
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.utils import docstring
from tomwer.core.process.reconstruction.nabu.utils import check_sinogram_half
from typing import Iterable, Union
from typing import Optional
from nabu.pipeline.config import generate_nabu_configfile
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from nabu.pipeline.config import _extract_nabuconfig_keyvals
from nabu import version as nabu_version
from silx.io.dictdump import h5todict
from . import utils
from . import settings as nabu_settings
from tomoscan.io import HDF5File
from tomwer.core.scan.edfscan import EDFTomoScan
from silx.io.utils import h5py_read_dataset
from nabu.pipeline.config import get_default_nabu_config
from tomwer.io.utils.h5pyutils import EntryReader
from tomwer.core.process.task import Task
from .target import Target
from .nabucommon import (
    _NabuBaseReconstructor,
    ResultsRun,
    ResultsLocalRun,
    ResultSlurmRun,
)
from processview.core.superviseprocess import SuperviseProcess

_logger = logging.getLogger(__name__)
if not has_nabu:
    _logger.error(_err_import_nabu)


def run_slices_reconstruction(
    scan: TomwerScanBase,
    config: dict,
    dry_run: bool = False,
    advancement=None,
    process_id=None,
) -> tuple:
    """
    call nabu for a reconstruction on scan with the given configuration

    :param TomwerScanBase scan: scan to reconstruct
    :param dict config: configuration to run the reconstruction.
                        Contains nabu reconstruction parameters and slurm cluster
                        configruation if requested (key: `slurm-cluster`).
    :param bool dry_run: do we want to run dry
    :param bool local: do we want to run a local reconstruction
    :param stderr: file to redirect stderr
    :param stdout: file to redirect stdout
    :param advancement: optional Progress class to display advancement
    :param int process_id: optional process id

    :return: (all_succeed, stdouts, stderrs, final_configs, future_scan)
             * all_succeed: bool, True if all the reconstruction succeed or if all job request succeed.
             * stdouts: list of stdout of job reconstruction or job requests
             * stderrs: list of stderr of job reconstruction or job requests
             * final_configs: list of configurations submits to nabu
             * future_scan: Optional[FutureTomwerScan] future scan containing futures pointing to job submited to the cluster.
                            None if local reconstruction
    :rtype: tuple

    Behavior: will clear the last slices reconstructed
    """
    # TODO: remove the local parameter
    check_sinogram_half(scan)
    _logger.info("start reconstruction of {}".format(str(scan)))

    cluster_config = config.pop("cluster_config", None)
    if cluster_config == {}:
        cluster_config = None
    is_cluster_job = cluster_config is not None
    if is_cluster_job and not is_slurm_available():
        raise ValueError(
            "job on cluster requested but no access to slurm cluster found"
        )

    # if scan contains some center of position copy it to nabu
    if scan.axis_params is not None and scan.axis_params.relative_cor_value is not None:
        if "reconstruction" in config:
            # move the cor value to the nabu reference
            cor_nabu_ref = scan.axis_params.relative_cor_value + scan.dim_1 / 2.0
            config["reconstruction"]["rotation_axis_position"] = str(cor_nabu_ref)
    _logger.info("set nabu reconstruction parameters to {}".format(str(scan)))

    # update nabu recons_params used
    sc_config = get_default_nabu_config(nabu_fullfield_default_config)
    sc_config.update(config)
    scan.nabu_recons_params = sc_config

    # handle special cases like several db...
    nabu_configurations = interpret_tomwer_configuration(config, scan=scan)
    if "phase" in config and "delta_beta" in config["phase"]:
        db_list = utils.retrieve_lst_of_value_from_str(
            config["phase"]["delta_beta"], type_=float
        )
    else:
        db_list = ()
    ask_sinogram_registration = len(db_list) > 0
    ask_sinogram_load = False
    output_urls = []
    stderrs = []
    stdouts = []
    final_configs = []
    futures = []
    clients = []
    all_succeed = True
    if advancement is not None:
        advancement.setMaxAdvancement(len(nabu_configurations))
    scan.clear_latest_reconstructions()
    for nabu_configuration in nabu_configurations:
        l_config, slice_index = nabu_configuration
        result = run_single_slice_reconstruction(
            nabu_config=l_config,
            cluster_config=cluster_config,
            scan=scan,
            slice_index=slice_index,
            dry_run=dry_run,
            ask_sinogram_registration=ask_sinogram_registration,
            ask_sinogram_load=ask_sinogram_load,
        )
        if slice_index is None:
            continue

        # specific treatments of results
        if result is None:
            # in case of timeout or another issue. Log should already have been provided
            pass
        elif isinstance(result, ResultsLocalRun):
            assert not is_cluster_job, "cluster job should not return ResultsLocalRun"
            stderrs.append(result.std_err)
            stdouts.append(result.std_out)
            output_urls.extend(result.results_urls)
            # if slice_index is None this mean that we are simply creating the
            # .cfg file for nabu full volume.
        elif isinstance(result, ResultSlurmRun):
            assert (
                is_cluster_job
            ), "local reconstruction should not return ResultSlurmRun"
            stderrs.append(result.std_err)
            stdouts.append(result.std_out)
            futures.extend(result.future_slurm_jobs)
            clients.append(result.client)
        elif not isinstance(result, ResultsRun):
            raise ValueError(
                f"result is expected to be an instance of {ResultsRun} not {type(result)}"
            )

        # common treatments of results
        if result is not None:
            final_configs.append(result.config)
            all_succeed = all_succeed and result.success

        if ask_sinogram_registration:
            ask_sinogram_load = True
            ask_sinogram_registration = False
        if advancement is not None:
            advancement.increaseAdvancement(1)

    if is_cluster_job:
        future_scan = FutureTomwerScan(
            scan=scan,
            futures=tuple(futures),
            clients=clients,
            process_requester_id=process_id,
        )
        return all_succeed, stdouts, stderrs, final_configs, future_scan
    else:
        # tag latest reconstructions
        scan.set_latest_reconstructions(output_urls)
        return all_succeed, stdouts, stderrs, final_configs, None


class NabuSlices(
    Task, SuperviseProcess, input_names=("data",), output_names=("data", "nabu_params")
):
    """Definition of the nabu reconstruction Single process"""

    def __init__(self, process_id=None, varinfo=None, inputs=None, label=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        Task.__init__(self, varinfo=varinfo, inputs=inputs, label=label)
        if "recons_params" in inputs:
            raise KeyError("Do not use recons_params but `nabu_recons_params` instead")
        recons_params = inputs.get("nabu_params", {})
        self.set_configuration(recons_params)
        self._dry_run = inputs.get("dry_run", False)

    def run(self):
        scan = self.inputs.data
        self.outputs.nabu_params = None
        if scan is None:
            self.outputs.data = None
            return
        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(
                "input type of {}: {} is not managed" "".format(scan, type(scan))
            )
        assert self.get_configuration() is not None, "configuration should be set"
        run_slices_reconstruction(
            scan=scan,
            config=self.get_configuration(),
            dry_run=self.dry_run,
            process_id=self.process_id,
        )
        # register result
        entry = "entry"
        if isinstance(scan, HDF5TomoScan):
            entry = scan.entry

        with scan.acquire_process_file_lock():
            self.register_process(
                process_file=scan.process_file,
                entry=entry,
                configuration=self.get_configuration(),
                results={},
                process_index=scan.pop_process_index(),
                overwrite=True,
            )
        self.outputs.data = scan
        self.outputs.nabu_params = scan.nabu_recons_params

    def set_configuration(self, configuration: dict) -> None:
        Task.set_configuration(self, configuration=configuration)
        if "dry_run" in configuration:
            self.set_dry_run(bool(configuration["dry_run"]))

    def pathReceived(self, scan):
        return self.process(scan=scan)

    # TODO: those get / set configuration should be removed now that Task is used
    def get_configuration(self):
        return self._settings

    @staticmethod
    def program_name():
        return "nabu-slices"

    @staticmethod
    def program_version():
        return nabu_version

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    @staticmethod
    def get_process_frm_process_file(process_file, entry):
        """
        Read informations regarding the nabu process save in the
        tomwer_process.h5 file

        :param process_file:
        :param entry:
        :return: dictionary with the contain of the nabu process
        :rtype:dict
        """
        if entry is None:
            with HDF5File(process_file, "r", swmr=True) as h5f:
                entries = NabuSlices._get_process_nodes(
                    root_node=h5f, process=NabuSlices
                )
                if len(entries) == 0:
                    _logger.info("unable to find a Axis process in %s" % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError("several entry found, entry should be " "specify")
                else:
                    entry = list(entries.keys())[0]
                    _logger.info("take %s as default entry" % entry)

        configuration_path = None
        res = {}

        with HDF5File(process_file, "r", swmr=True) as h5f:
            nabu_nodes = NabuSlices._get_process_nodes(
                root_node=h5f[entry], process=NabuSlices
            )
            index_to_path = {}
            for key, index in nabu_nodes.items():
                index_to_path[index] = key

            if len(nabu_nodes) == 0:
                return None
            # take the last processed dark ref
            last_process_index = sorted(list(nabu_nodes.values()))[-1]
            last_process_dark = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug(
                    "several processing found for dark-ref, "
                    "take the last one: %s" % last_process_dark
                )

            for key_name in (
                "class_instance",
                "date",
                "program",
                "sequence_index",
                "version",
            ):
                if key_name in h5f[last_process_dark]:
                    res[key_name] = h5py_read_dataset(h5f[last_process_dark][key_name])
            if "configuration" in h5f[last_process_dark]:
                configuration_path = "/".join(
                    (h5f[last_process_dark].name, "configuration")
                )

        if configuration_path is not None:
            res["configuration"] = h5todict(
                h5file=process_file, path=configuration_path
            )
        return res

    @staticmethod
    def retrieve_last_relative_cor(scan):
        with EntryReader(scan.process_file_url) as h5f:
            latest_nabu_node = Task.get_most_recent_process(h5f, NabuSlices)
            path = "configuration/reconstruction/rotation_axis_position"
            if latest_nabu_node is not None and path in latest_nabu_node:
                return h5py_read_dataset(latest_nabu_node[path])


def interpret_tomwer_configuration(
    config: dict, scan: Union[TomwerScanBase, None]
) -> tuple:
    """
    tomwer can 'mock' the nabu reconstruction to request more feature.
    Typical use case is that we can ask for reconstruction of several
    slices and not only the volume

    :param dict config: tomwer configuration for nabu
    :param scan: requested if we want to get slices
    :return: tuple of tuples (nabu configuration, is slice)
    """
    if not isinstance(config, dict):
        raise TypeError("config is expected to be a dict")

    def get_nabu_config(config):
        nabu_config = copy.deepcopy(config)
        if "tomwer_slices" in nabu_config:
            del nabu_config["tomwer_slices"]
        return nabu_config

    if "tomwer_slices" in config and scan is not None:
        slices = list(NabuSliceMode.getSlices(config["tomwer_slices"], scan))
    else:
        slices = []

    if "phase" in config and "delta_beta" in config["phase"]:
        pag_dbs = utils.retrieve_lst_of_value_from_str(
            config["phase"]["delta_beta"], type_=float
        )
        if len(pag_dbs) == 0:
            pag_dbs = (None,)
    else:
        pag_dbs = (None,)

    # by default add the slice 'None' which is the slice for the volume
    slices.append(None)
    nabu_config = get_nabu_config(config=config)
    res = []
    for slice_ in slices:
        for pag_db in pag_dbs:
            local_config = copy.deepcopy(nabu_config)
            if pag_db is not None:
                local_config["phase"]["delta_beta"] = str(pag_db)
            res.append((local_config, slice_))
    return tuple(res)


class SingleSliceRunner(_NabuBaseReconstructor):
    def __init__(
        self,
        scan: TomwerScanBase,
        config: dict,
        dry_run: bool,
        slice_index: Union[int, str, None],
        target: Target,
        ask_sinogram_registration: bool,
        ask_sinogram_load: bool,
        cluster_config: Optional[dict],
        process_name: str,
        add_to_latest_reconstructions: bool = True,
    ) -> None:
        super().__init__(
            scan=scan,
            dry_run=dry_run,
            target=target,
            cluster_config=cluster_config,
            process_name=process_name,
        )
        self._slice_index = slice_index
        if not isinstance(config, dict):
            raise TypeError(f"config is expected to be a dictionary not {type(dict)}")
        self._config = config
        self._ask_sinogram_registration = ask_sinogram_registration
        self._ask_sinogram_load = ask_sinogram_load
        self._add_to_latest_reconstructions = add_to_latest_reconstructions

    @property
    def slice_index(self):
        return self._slice_index

    @property
    def config(self):
        return self._config

    @property
    def ask_sinogram_load(self):
        return self._ask_sinogram_load

    @property
    def ask_sinogram_registration(self):
        return self._ask_sinogram_registration

    @property
    def add_to_latest_reconstructions(self):
        return self._add_to_latest_reconstructions

    @docstring(_NabuBaseReconstructor)
    def only_create_config_file(self):
        return self.slice_index is None

    @docstring(_NabuBaseReconstructor)
    def run(self) -> Iterable:
        """
        If the target is local will wait for the reconstruction to be finish to return

        :raise: TIMEOUT_SLURM_JOB_SUBMISSION if not all workers spwan
        """
        if isinstance(self.slice_index, str):
            if self.slice_index == "middle":
                self._slice_index = self.scan.dim_2 // 2
            else:
                raise ValueError(
                    "slice index is expected to an int or 'middle' "
                    "and not {}".format(type(self.slice_index))
                )
        elif not isinstance(self.slice_index, (int, type(None))):
            raise TypeError(
                "slice index is expected to an int or 'middle' or None"
                "and not {}".format(type(self.slice_index))
            )
        config_complete = _extract_nabuconfig_keyvals(nabu_fullfield_default_config)
        config_complete["dataset"] = self.scan.get_nabu_dataset_info()
        for key in config_complete.keys():
            if key in self.config:
                config_complete[key].update(self.config[key])
        config = config_complete

        config["resources"] = utils.get_nabu_resources_desc(
            scan=self.scan, workers=1, method="local"
        )
        # force overwrite results
        if "output" not in config:
            config["output"] = {}
        # handle nabu sinogram step
        if self.scan.process_file is not None:
            steps_file_basename, _ = os.path.splitext(self.scan.process_file)
            steps_file_basename = "_".join(
                ("steps_file_basename", "nabu", "sinogram", "save", "step")
            )
            steps_file_basename = steps_file_basename + ".hdf5"
            steps_file = os.path.join(
                os.path.dirname(self.scan.process_file), steps_file_basename
            )
        else:
            steps_file = ""

        config["output"].update({"overwrite_results": 1})
        config["pipeline"] = {
            "save_steps": "sinogram" if self.ask_sinogram_registration else "",
            "resume_from_step": "sinogram" if self.ask_sinogram_load else "",
            "steps_file": steps_file,
        }

        config = self._treateOutputSliceConfig(config)
        # the policy is to save nabu .cfg file at the same location as the
        # force overwrite results
        if self.slice_index is not None:
            config["reconstruction"]["start_z"] = self.slice_index
            config["reconstruction"]["end_z"] = self.slice_index
        cfg_folder = os.path.join(
            config["output"]["location"], nabu_settings.NABU_CFG_FILE_FOLDER
        )
        if not os.path.exists(cfg_folder):
            os.makedirs(cfg_folder)

        name = (
            config["output"]["file_prefix"] + nabu_settings.NABU_CONFIG_FILE_EXTENSION
        )
        if not isinstance(self.scan, EDFTomoScan):
            name = "_".join((self.scan.entry, name))
        conf_file = os.path.join(cfg_folder, name)
        _logger.info("{}: create {}".format(self.scan, conf_file))
        # add some tomwer metadata and save the configuration
        # note: for now the section is ignored by nabu but shouldn't stay that way
        with utils.TomwerInfo(config) as config_to_dump:
            generate_nabu_configfile(
                conf_file,
                nabu_fullfield_default_config,
                config=config_to_dump,
                options_level="advanced",
            )

        return tuple(
            [
                self._process_config(
                    config_to_dump=config_to_dump,
                    config_file=conf_file,
                    start_z=None,
                    end_z=None,
                    info="nabu slice reconstruction",
                    file_format=config_to_dump["output"]["file_format"],
                    process_name=self.process_name,
                ),
            ]
        )

    @docstring(_NabuBaseReconstructor)
    def _get_futures_slurm_callback(self, config_to_dump):
        if self.add_to_latest_reconstructions:
            # add callback to set slices reconstructed urls
            class CallBack:
                # we cannot create a future directly because distributed enforce
                # the callback to have a function signature with only the future
                # as single parameter.
                def __init__(self, f_partial, scan) -> None:
                    self.f_partial = f_partial
                    self.scan = scan

                def process(self, fn):
                    if fn.done() and not (fn.cancelled() or fn.exception()):
                        # update reconstruction urls only if processing succeed.
                        recons_urls = self.f_partial()
                        self.scan.add_latest_reconstructions(recons_urls)

            file_format = config_to_dump["output"]["file_format"]
            callback = functools.partial(
                utils.get_recons_urls,
                file_prefix=config_to_dump["output"]["file_prefix"],
                location=config_to_dump["output"]["location"],
                file_format=file_format,
                scan=self.scan,
                slice_index=None,
                start_z=None,  # self.slice_index,
                end_z=None,  # self.slice_index,
            )

            return (CallBack(callback, self.scan),)
        else:
            return super()._get_futures_slurm_callback(config_to_dump)

    @staticmethod
    def get_file_basename_reconstruction(scan, pag, db, slice_index: Union[str, int]):
        if pag:
            assert db is not None, "if paganin defined, db should not be None"
        if slice_index == "middle":
            slice_index = scan.dim_2 // 2
        assert type(db) in (int, type(None))
        if isinstance(scan, HDF5TomoScan):
            basename, _ = os.path.splitext(scan.master_file)
            basename = os.path.basename(basename)
        else:
            basename = os.path.basename(scan.path)
        if slice_index is None:
            if pag:
                return "_".join((basename + "pag", "db" + str(db).zfill(4)))
            else:
                return basename
        else:
            if pag:
                return "_".join(
                    (
                        basename + "slice_pag",
                        str(slice_index).zfill(4),
                        "db" + str(db).zfill(4),
                    )
                )
            else:
                return "_".join((basename + "slice", str(slice_index).zfill(4)))

    @docstring(_NabuBaseReconstructor)
    def _get_file_basename_reconstruction(self, pag, db):
        """

        :param TomwerScanBase scan: scan reconstructed
        :param Union[None, int] slice_index: index of the slice reconstructed.
                                            if None, we want to reconstruct the
                                            entire volume
        :param bool pag: is it a paganin reconstruction
        :param int db: delta / beta parameter
        :return: basename of the file reconstructed (without any extension)
        """
        return self.get_file_basename_reconstruction(
            scan=self.scan, db=db, pag=pag, slice_index=self.slice_index
        )


def run_single_slice_reconstruction(
    scan,
    nabu_config,
    dry_run,
    slice_index: Union[int, str, None],
    ask_sinogram_registration: bool,
    ask_sinogram_load: bool,
    process_id: Optional[int] = None,
    cluster_config: Optional[dict] = None,
    add_to_latest_reconstructions=True,
) -> Optional[ResultsRun]:
    """
    # TODO: might need something like a context or an option "keep" slice in memory

    :param scan:
    :param nabu_config: configruation of nabu process
    :param cluster_config: configruation of cluster (slurm-cluster only for now)
    :param dry_run:
    :param Union[None,int, str] slice_index: slice index to reconstruct.
                                             If str should be "middle"
    :param local:
    :param stdout: file to redirect stdout
    :param stderr: file to redirect stderr
    :param bool ask_sinogram_registration: should we ask nabu to register sinogram
    :param bool ask_sinogram_load: should we ask nabu to load sinogram
    :param bool add_to_latest_reconstructions: if true add reconstructed slice to the latest reconstruction.
                                               We wan't to avoid this treatment for saaxis and sadeltebeta for example
    :return: result of the slice reconstruction if succeed to launch it.
    :rtype: Optional[ResultsRun]
    """
    # TODO: remove local from the function signature
    target = Target.SLURM if cluster_config not in ({}, None) else Target.LOCAL

    if process_id is not None:
        try:
            process_name = ProcessManager().get_process(process_id=process_id).name
        except KeyError:
            process_name = "unknow"
    else:
        process_name = ""

    slice_reconstructor = SingleSliceRunner(
        scan=scan,
        config=nabu_config,
        dry_run=dry_run,
        slice_index=slice_index,
        target=target,
        ask_sinogram_registration=ask_sinogram_registration,
        ask_sinogram_load=ask_sinogram_load,
        cluster_config=cluster_config,
        add_to_latest_reconstructions=add_to_latest_reconstructions,
        process_name=process_name,
    )
    try:
        results = slice_reconstructor.run()
    except TimeoutError as e:
        _logger.error(e)
        return None
    else:
        assert len(results) == 1, "only one slice should be reconstructed"
        return results[0]


class NabuSliceMode(_Enum):
    MIDDLE = "middle"
    OTHER = "other"

    @staticmethod
    def getSlices(slices, scan) -> tuple:
        res = []
        try:
            mode = NabuSliceMode.from_value(slices)
        except ValueError:
            try:
                res = utils.retrieve_lst_of_value_from_str(slices, type_=int)
            except Exception:
                pass
        else:
            if mode == mode.MIDDLE:
                n_slice = scan.dim_2 or 2048
                res.append(n_slice // 2)
            else:
                raise ValueError(
                    "there should be only two ways of defining "
                    "slices: middle one or other, by giving "
                    "an unique value or a list or a tuple"
                )
        return tuple(res)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/06/2021"


import unittest
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.process.reconstruction.ftseries.params.fastsetupdefineglobals import (
    FastSetupAll,
)
from tomwer.core.utils.pyhstutils import _findPyHSTVersions, _getPyHSTDir
from tomwer.gui.reconstruction.ftserie import FtserieWidget
from tomwer.gui.qtapplicationmanager import QApplicationManager
from tomwer.gui.reconstruction.recpyhstwidget import RecPyHSTWidget
from tomwer.synctools.ftseries import QReconsParams
from tomwer.test.utils import skip_gui_test

pyhstVersion = _findPyHSTVersions(_getPyHSTDir())
_qapp = QApplicationManager()


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestRecPyHSTWidget(TestCaseQt):
    """Make sure the gui is correctly editing the ReconsParam and the
    RecPyHST class
    """

    def setUp(self):
        self._recons_params = QReconsParams()
        self.ftserieWidget = FtserieWidget(recons_params=self._recons_params)
        self.ftserieWidget.setAttribute(qt.Qt.WA_DeleteOnClose)

        self.widget = RecPyHSTWidget(
            parent=None, recons_params=self._recons_params.pyhst
        )
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        assert id(self.ftserieWidget.recons_params.pyhst) == id(
            self._recons_params.pyhst
        )
        assert id(self.widget.recons_params) == id(self._recons_params.pyhst)

    def tearDown(self):
        self.widget.close()
        self.ftserieWidget.close()

    @pytest.mark.skipif(len(pyhstVersion) == 0, reason="PyHST2 missing")
    def testUpdating(self):
        """Check behavior of the gui when editing the self.recons_params
        Make sure also that sync with the ftserie is made
        """
        pyhstWidget = self.ftserieWidget.getReconsParamSetEditor()._PyHSTWidget

        # Test make oar
        isMakeOAR = self.widget.recons_params["MAKE_OAR_FILE"]
        self.assertTrue(self.widget._makeOARFileCB.isChecked() == isMakeOAR)
        self.assertTrue(pyhstWidget._makeOARFileCB.isChecked() == isMakeOAR)
        self.widget.recons_params["MAKE_OAR_FILE"] = not isMakeOAR
        self.assertTrue(self.widget._makeOARFileCB.isChecked() != isMakeOAR)
        self.assertTrue(pyhstWidget._makeOARFileCB.isChecked() != isMakeOAR)

        # test pyhst version
        self.assertTrue(self.widget._qcbPyHSTVersion.currentText() == FastSetupAll.OFFV)
        self.assertTrue(pyhstWidget._qcbPyHSTVersion.currentText() == FastSetupAll.OFFV)
        self._recons_params["PYHSTEXE"]["EXE"] = "toto"
        self.assertTrue(pyhstWidget._qcbPyHSTVersion.currentText() == "toto")
        self.assertTrue(self.widget._qcbPyHSTVersion.currentText() == "toto")

    @pytest.mark.skipif(len(pyhstVersion) == 0, reason="PyHST2 missing")
    def testEdition(self):
        """Check behavior of ReconsParam when editing i through the gui"""
        pyhstWidget = self.ftserieWidget.getReconsParamSetEditor()._PyHSTWidget

        # Test make oar
        oVal = self.widget._makeOARFileCB.isChecked()
        self.assertTrue(oVal == pyhstWidget._makeOARFileCB.isChecked())
        self.widget._makeOARFileCB.setChecked(not oVal)
        self.qapp.processEvents()

        self.assertTrue(self.widget.recons_params["MAKE_OAR_FILE"] != oVal)
        self.assertFalse(oVal == pyhstWidget._makeOARFileCB.isChecked())
        pyhstWidget._makeOARFileCB.setChecked(oVal)
        self.assertTrue(self.widget._makeOARFileCB.isChecked() == oVal)
        self.assertTrue(self.widget.recons_params["MAKE_OAR_FILE"] == oVal)

        # test pyhst version
        oExe = self.widget.recons_params["EXE"] = "toto"
        self.assertTrue(self.widget._qcbPyHSTVersion.currentText() == oExe)
        self.assertTrue(pyhstWidget._qcbPyHSTVersion.currentText() == oExe)
        self.widget._qcbPyHSTVersion.addItem("toto")
        iToto = self.widget._qcbPyHSTVersion.findText("toto")
        self.assertTrue(iToto > 0)
        self.widget._qcbPyHSTVersion.setCurrentIndex(iToto)
        self.assertTrue(self.widget._qcbPyHSTVersion.currentText() == "toto")
        self.assertTrue(pyhstWidget._qcbPyHSTVersion.currentText() == "toto")
        self.assertTrue(self.widget.recons_params["EXE"] == "toto")
        iOExe = self.widget._qcbPyHSTVersion.findText(oExe)
        self.assertTrue(iOExe > 0)
        pyhstWidget._qcbPyHSTVersion.setCurrentIndex(iOExe)
        self.assertTrue(self.widget._qcbPyHSTVersion.currentText() == oExe)
        self.assertTrue(pyhstWidget._qcbPyHSTVersion.currentText() == oExe)
        self.assertTrue(self.widget.recons_params["EXE"] == oExe)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestRecPyHSTWidget,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain the SADeltaBetaProcess. Half automatic best delta / beta finder
"""

__authors__ = [
    "H.Payno",
]
__license__ = "MIT"
__date__ = "28/10/2021"


from copy import copy
from tomwer.core.process.reconstruction.nabu.nabucommon import (
    ResultSlurmRun,
    ResultsLocalRun,
    ResultsWithStd,
)
from tomwer.core.scan.scanfactory import ScanFactory
from .params import SADeltaBetaParams
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils import logconfig
from tomwer.core.process.reconstruction.nabu.utils import check_sinogram_half
from processview.core.manager import ProcessManager, DatasetState
from processview.core.superviseprocess import SuperviseProcess
from tomwer.core.process.task import Task
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomoscan.io import HDF5File
from silx.io.utils import get_data
import tomwer.version
from tomwer.core.process.reconstruction.nabu.nabuslices import (
    run_single_slice_reconstruction,
)
from tomwer.core.process.reconstruction.nabu.nabucommon import ResultsRun
from tomwer.core.progress import Progress
from typing import Optional, Union
import os
import logging
import h5py
import numpy
from ..nabu import utils as nabu_utils
from tomwer.core.process.reconstruction.scores import compute_score
from tomwer.core.process.reconstruction.scores import ScoreMethod
from tomwer.core.process.reconstruction.scores import ComputedScore
from tomwer.core.process.reconstruction.scores import get_disk_mask_radius, apply_roi
from tomwer.core.process.reconstruction.nabu.nabuslices import SingleSliceRunner

_logger = logging.getLogger(__name__)


DEFAULT_RECONS_FOLDER = "sadeltabeta_results"


def one_slice_several_db(
    scan: TomwerScanBase,
    configuration: Union[dict, SADeltaBetaParams],
    process_id: Optional[int] = None,
) -> tuple:
    """
    Run a slice reconstruction using nabu per Center Of Rotation (cor) provided
    Then for each compute a score (quality) of the center of rotation

    :param TomwerScanBase scan:
    :param Union[dict,SADeltaBetaParams] configuration:
    :return: cor_reconstructions, outs, errs
             cor_reconstructions is a dictionary of cor as key and a tuple
             (url, score) as value
    :rtype: tuple
    """
    if isinstance(configuration, dict):
        configuration = SADeltaBetaParams.from_dict(configuration)
    elif not isinstance(configuration, SADeltaBetaParams):
        raise TypeError(
            "configuration should be a dictionary or an instance of SAAxisParams"
        )
    check_sinogram_half(scan)

    if scan.axis_params is None:
        from tomwer.core.process.reconstruction.axis import AxisRP

        scan.axis_params = AxisRP()

    configuration.check_configuration()
    slice_index = configuration.slice_indexes
    delta_beta_s = configuration.delta_beta_values
    output_dir = configuration.output_dir
    dry_run = configuration.dry_run
    cluster_config = configuration.cluster_config
    _logger.info(
        "launch reconstruction of slice {} and delta / beta: {}".format(
            slice_index, delta_beta_s
        )
    )
    if isinstance(slice_index, str):
        if not slice_index == "middle":
            raise ValueError("slice index {} not recognized".format(slice_index))
    elif not len(slice_index) == 1:
        raise ValueError("only manage one slice")
    else:
        slice_index = list(slice_index.values())[0]
    advancement = Progress("sa-delta-beta - slice {} of {}".format(slice_index, scan))

    config = configuration.nabu_params.copy()

    check_sinogram_half(scan)
    _logger.info("start reconstruction of {}".format(str(scan)))
    # if scan contains some center of position copy it to nabu
    if scan.axis_params is not None and scan.axis_params.relative_cor_value is not None:
        if "reconstruction" in config:
            # move the cor value to the nabu reference
            cor_nabu_ref = scan.axis_params.relative_cor_value + scan.dim_1 / 2.0
            config["reconstruction"]["rotation_axis_position"] = str(cor_nabu_ref)

    _logger.info("set nabu reconstruction parameters to {}".format(str(scan)))
    scan.nabu_recons_params = config

    db_reconstructions = {}
    # key is delta / beta, value is url
    stderrs = []
    stdouts = []
    all_succeed = True
    future_scans = {}
    # key is delta / beta, value is future
    if advancement is not None:
        advancement.setMaxAdvancement(len(delta_beta_s))
    for db in delta_beta_s:
        if "output" not in config:
            config["output"] = {}
        if output_dir is None:
            config["output"]["location"] = os.path.join(
                scan.path, DEFAULT_RECONS_FOLDER
            )
        else:
            config["output"]["location"] = output_dir
        # TODO: allow file format modifications
        config["output"]["file_format"] = "hdf5"
        if "phase" not in config:
            config["phase"] = {}
        config["phase"]["delta_beta"] = db
        config["phase"]["method"] = "Paganin"
        res = run_single_slice_reconstruction(
            nabu_config=config,
            scan=scan,
            slice_index=slice_index,
            dry_run=dry_run,
            ask_sinogram_registration=False,
            ask_sinogram_load=False,
            cluster_config=cluster_config,
            add_to_latest_reconstructions=False,
            process_id=process_id,
        )
        if isinstance(res, ResultsRun):
            all_succeed = all_succeed and res.success
        if isinstance(res, ResultSlurmRun):
            future_scans[db] = res.future_slurm_jobs
        if isinstance(res, ResultsWithStd):
            if slice_index is not None:
                stderrs.append(res.std_err)
                stdouts.append(res.std_out)
        if isinstance(res, ResultsLocalRun) and len(res.results_urls) > 0:
            assert len(res.results_urls) == 1, "only one slice expected"
            db_reconstructions[db] = res.results_urls[0]

        if advancement is not None:
            advancement.increaseAdvancement(1)

    pag = False
    if "phase" in config:
        if "method" in config["phase"] and config["phase"]["method"] != "":
            pag = True

    # treat future.
    for db, future_scan_list in future_scans.items():
        assert (
            len(future_scan_list) == 1
        ), "only one future should be created for one slice / db couple"
        future = future_scan_list[0]
        future.result()
        if future.cancelled() or future.exception():
            continue
        file_prefix = SingleSliceRunner.get_file_basename_reconstruction(
            scan=scan,
            slice_index=slice_index,
            pag=pag,
            db=int(db) if db is not None else None,
        )
        # retrieve url
        recons_url = nabu_utils.get_recons_urls(
            file_prefix=file_prefix,
            location=config["output"]["location"],
            file_format=config.get("file_format", "hdf5"),
            scan=scan,
            slice_index=None,
            start_z=None,
            end_z=None,
        )

        assert len(recons_url) <= 1, "only one slice expected"
        if len(recons_url) == 1:
            db_reconstructions[db] = recons_url[0]
        else:
            _logger.warning(
                f"something went wrong with reconstruction of {db} from {str(scan)}"
            )

    class PostProcessing:
        def run(self):
            datasets = self.load_datasets()

            mask_disk_radius = get_disk_mask_radius(datasets)
            scores = {}
            rois = {}
            for db, (url, data) in datasets.items():
                if data is None:
                    score = None
                else:
                    assert data.ndim == 2
                    data_roi = apply_roi(data=data, radius=mask_disk_radius, url=url)
                    rois[db] = data_roi

                    # move data_roi to [0-1] range
                    #  preprocessing: get percentile 0 and 99 from image and
                    #  "clean" highest and lowest pixels from it
                    min_p, max_p = numpy.percentile(data_roi, (1, 99))
                    data_roi_int = data_roi[...]
                    data_roi_int[data_roi_int < min_p] = min_p
                    data_roi_int[data_roi_int > max_p] = max_p
                    data_roi_int = (data_roi_int - min_p) / (max_p - min_p)

                    score = ComputedScore(
                        tv=compute_score(data=data_roi_int, method=ScoreMethod.TV),
                        std=compute_score(data=data_roi_int, method=ScoreMethod.STD),
                    )
                scores[db] = (url, score)
            return scores, rois

        def load_datasets(self):
            datasets_ = {}
            for db, url in db_reconstructions.items():
                try:
                    data = get_data(url=url)
                except Exception as e:
                    _logger.error(
                        "Fail to compute a score for {}. Reason is {}"
                        "".format(url.path(), str(e))
                    )
                    datasets_[db] = (url, None)
                else:
                    if data.ndim == 3:
                        if data.shape[0] == 1:
                            data = data.reshape(data.shape[1], data.shape[2])
                        elif data.shape[2] == 1:
                            data = data.reshape(data.shape[0], data.shape[1])
                        else:
                            raise ValueError(
                                "Data is expected to be 2D. Not {}".format(data.ndim)
                            )
                    elif data.ndim == 2:
                        pass
                    else:
                        raise ValueError(
                            "Data is expected to be 2D. Not {}".format(data.ndim)
                        )

                    datasets_[db] = (url, data)
            return datasets_

    post_processing = PostProcessing()
    scores, rois = post_processing.run()
    return scores, stdouts, stderrs, rois


class SADeltaBetaProcess(
    Task, SuperviseProcess, input_names=("data",), output_names=("data",)
):
    """
    Main process to launch several reconstruction of a single slice with
    several Center Of Rotation (cor) values
    """

    def __init__(self, process_id=None, varinfo=None, inputs=None, label=None):
        Task.__init__(self, varinfo=varinfo, inputs=inputs, label=label)
        SuperviseProcess.__init__(self, process_id=process_id)
        self._dry_run = inputs.get("dry_run", False)
        self._dump_process = inputs.get("dump_process", True)
        self._dump_roi = inputs.get("dump_roi", False)
        self._sa_delta_beta_params = inputs.get("sa_delta_beta_params", None)
        if self._sa_delta_beta_params is not None:
            self.set_configuration(properties=self._sa_delta_beta_params)
        self._std_outs = tuple()
        self._std_errs = tuple()

    @property
    def dump_roi(self):
        return self._dump_roi

    @dump_roi.setter
    def dump_roi(self, dump):
        self._dump_roi = dump

    @property
    def std_outs(self):
        return self._std_outs

    @property
    def std_errs(self):
        return self._std_errs

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    def set_configuration(self, configuration: dict) -> None:
        if isinstance(configuration, SADeltaBetaParams):
            self._settings = configuration.to_dict()
        elif isinstance(configuration, dict):
            self._settings = configuration
        else:
            raise TypeError(
                "configuration should be an instance of dict or " "SAAxisParams"
            )

    @staticmethod
    def autofocus(scan):
        scores = scan.sa_delta_beta_params.scores
        if scores is None:
            return
        score_method = scan.sa_delta_beta_params.score_method
        best_db, best_score = None, 0
        for cor, (_, score_cls) in scores.items():
            score = score_cls.get(score_method)
            if score is None:
                continue
            if score > best_score:
                best_db, best_score = cor, score
        scan.sa_delta_beta_params.autofocus = best_db
        scan.sa_delta_beta_params.value = best_db

    def run(self):
        scan = self.inputs.data
        if scan is None:
            self.outputs.data = scan
            return
        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(
                "input type of {}: {} is not managed" "".format(scan, type(scan))
            )
        # insure scan contains some parameter regarding sa delta / beta
        if scan.sa_delta_beta_params is None:
            scan.sa_delta_beta_params = SADeltaBetaParams()

        # create dir if does not exists
        config = copy(self.get_configuration())
        params = SADeltaBetaParams.from_dict(config)
        if params.output_dir in (None, ""):
            params.output_dir = os.path.join(scan.path, DEFAULT_RECONS_FOLDER)
            if not os.path.exists(params.output_dir):
                os.makedirs(params.output_dir)
        db_res, self._std_outs, self._std_errs, rois = one_slice_several_db(
            scan=scan,
            configuration=params.to_dict(),
            process_id=self.process_id,
        )
        scan.sa_delta_beta_params.scores = db_res
        self.autofocus(scan=scan)
        self._process_end(scan=scan, db_res=db_res, score_rois=rois)
        self.outputs.data = scan

    def _process_end(self, scan, db_res, score_rois):
        assert isinstance(scan, TomwerScanBase)
        try:
            extra = {
                logconfig.DOC_TITLE: self._scheme_title,
                logconfig.SCAN_ID: str(scan),
            }
            slice_index = self.get_configuration().get("slice_index", None)

            if db_res is None:
                info = (
                    "fail to compute delta/beta scores of slice {} for scan {}."
                    "".format(slice_index, scan)
                )
                _logger.processFailed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.FAILED, details=info
                )
            else:
                info = "delta/beta scores of slice {} for scan {} computed." "".format(
                    slice_index, scan
                )
                _logger.processSucceed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan,
                    process=self,
                    state=DatasetState.WAIT_USER_VALIDATION,
                    details=info,
                )
        except Exception as e:
            _logger.error(e)
        else:
            if self._dump_process:
                process_idx = SADeltaBetaProcess.process_to_tomwer_processes(
                    scan=scan,
                )
                if self.dump_roi and process_idx is not None:
                    self.dump_rois(
                        scan, score_rois=score_rois, process_index=process_idx
                    )

    @staticmethod
    def dump_rois(scan, score_rois: dict, process_index: int):
        if score_rois is None or len(score_rois) == 0:
            return
        if not isinstance(score_rois, dict):
            raise TypeError("score_rois is expected to be a dict")
        process_file = scan.process_file
        process_name = "tomwer_process_" + str(process_index)

        if scan.saaxis_params.scores in (None, {}):
            return

        def get_process_path():
            return "/".join((scan.entry or "entry", process_name))

        # save it to the file
        with Task._get_lock(process_file):
            # needs an extra lock for multiprocessing

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                score_roi_grp = nx_process.require_group("score_roi")
                for db, roi in score_rois.items():
                    score_roi_grp[str(db)] = roi
                    score_roi_grp[str(db)].attrs["interpretation"] = "image"

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "semi-automatic delta/beta finder"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "Semi automatic center of rotation / axis calculation"

    @staticmethod
    def process_to_tomwer_processes(scan):
        if scan.process_file is not None:
            entry = "entry"
            if isinstance(scan, HDF5TomoScan):
                entry = scan.entry

            db = None
            if scan.sa_delta_beta_params is not None:
                db = scan.sa_delta_beta_params.selected_delta_beta_value

            process_index = scan.pop_process_index()
            try:
                with scan.acquire_process_file_lock():
                    Task._register_process(
                        process_file=scan.process_file,
                        entry=entry,
                        results={"delta_beta": db if db is not None else "-"},
                        configuration=scan.sa_delta_beta_params.to_dict(),
                        process_index=process_index,
                        overwrite=True,
                        process=SADeltaBetaProcess,
                    )
                    SADeltaBetaProcess._extends_results(
                        scan=scan, entry=entry, process_index=process_index
                    )
            except Exception as e:
                _logger.warning(
                    f"Fail to register process of with index {process_index}. Reason is {e}"
                )

            return process_index

    @staticmethod
    def _extends_results(scan, entry, process_index):
        process_file = scan.process_file
        process_name = "tomwer_process_" + str(process_index)

        if scan.sa_delta_beta_params.scores in (None, {}):
            return

        def get_process_path():
            return "/".join((entry or "entry", process_name))

        # save it to the file
        with Task._get_lock(process_file):
            # needs an extra lock for multiprocessing

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                if "NX_class" not in nx_process.attrs:
                    nx_process.attrs["NX_class"] = "NXprocess"

                results = nx_process.require_group("results")
                for cor, (url, score) in scan.sa_delta_beta_params.scores.items():
                    results_db = results.require_group(str(cor))
                    for method in ScoreMethod:
                        if method is ScoreMethod.TOMO_CONSISTENCY:
                            continue
                        results_db[method.value] = score.get(method)

                    link_path = os.path.relpath(
                        url.file_path(),
                        os.path.dirname(process_file),
                    )
                    results_db["reconstructed_slice"] = h5py.ExternalLink(
                        link_path, url.data_path()
                    )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui relative to semi-automatic axis calculation
"""


__authors__ = [
    "H. Payno",
]
__license__ = "MIT"
__date__ = "26/02/2021"


from tomwer.gui import icons
from tomwer.core.process.reconstruction.saaxis.saaxis import ReconstructionMode
from tomwer.gui.utils.slider import LogSlider
from tomwer.core.scan.scanbase import TomwerScanBase
from typing import Union
from silx.gui.plot import items
from silx.utils.enum import Enum as _Enum
from tomwer.gui.reconstruction.saaxis.sliceselector import SliceSelector
from tomwer.gui.visualization.sinogramviewer import SinogramViewer as _SinogramViewer
from tomwer.gui.utils.lineselector import QSliceSelectorDialog
from silx.gui import qt, icons as silxicons
from silx.gui.dialog.ImageFileDialog import ImageFileDialog
from silx.io.utils import get_data
import numpy
import weakref
import logging
import functools

_logger = logging.getLogger(__name__)


class DetectionAccuracy(_Enum):
    """The different accuracy modes"""

    COARSE = "Coarse"
    FINE = "Fine"
    ACCURATE = "Accurate"
    DEFINED_RANGE = "Defined range"


ACCURACIES = {
    DetectionAccuracy.COARSE: 500,
    DetectionAccuracy.FINE: 10,
    DetectionAccuracy.ACCURATE: 2,
}


class SinogramViewer(_SinogramViewer):
    """Redefine the main SinogramViewer for saaxis purpose:
    * options orientation is horizontal
    * add the display of the cor roi to be displayed
    * rename ApplyButton to load and change the icon
    * hide the line selection (control by the slice selector
    * move the dock widget to the bottom
    * rename subsampling "loaded sinogram subsampling" to provide more context
    """

    sigCorEdited = qt.Signal(float)
    """Signal emitted when the cor position change. Provided in absolute
    position"""
    sigWidthEdited = qt.Signal(float)
    """Signal emitted when the detection accuracy (width) change."""

    def __init__(self, parent):
        super().__init__(parent, opts_orientation=qt.Qt.Horizontal)
        self._fromMarker = None
        self._toMarker = None
        self._corMarker = None
        self._min = None
        self._max = None
        self._estimated_cor = None
        self._other_cors = None
        # hide line selection
        self._options.setLineSelectionVisible(False)
        # change ApplyButton name and icon
        self._loadButton = self._options._buttons.button(qt.QDialogButtonBox.Apply)
        self._loadButton.setText("load")
        style = qt.QApplication.style()
        self._loadButton.setIcon(style.standardIcon(qt.QStyle.SP_BrowserReload))
        # change subsampling name
        self._options._subsamplingLabel.setText("Sinogram subsampling")
        self._options._subsamplingLabel.setToolTip(
            "Define subsampling of the " "sinogram to be displayed"
        )

    def loadSinogram(self):
        """simulate the click on `load` of the QPushButton"""
        self._loadButton.clicked.emit()

    def getOptionsDockWidget(self):
        return self._dockOpt

    def _updatePlot(self, sinogram):
        super()._updatePlot(sinogram)
        if (
            self._min is not None
            and self._max is not None
            and self._estimated_cor is not None
        ):
            self.setCorRange(
                min=self._min,
                max=self._max,
                estimated_cor=self._estimated_cor,
                other_cors=self._other_cors,
            )

    def setCorRange(self, min, max, estimated_cor, other_cors=tuple()):
        """
        For now we can only modify the estimated cor

        :param float min:
        :param float max:
        :param float estimated_cor:
        :param int display_n_recons: the grid created by all cor observe
        :return:
        """
        self._min = min
        self._max = max
        self._estimated_cor = estimated_cor
        self._other_cors = other_cors if other_cors is not None else tuple()

        plot = self._plot.getPlotWidget()

        if self._corMarker is not None:
            self._corMarker.sigDragFinished.disconnect(self._middleMarkerMoved)
        if self._fromMarker is not None:
            self._fromMarker.sigDragFinished.disconnect(self._fromCallback)
        if self._toMarker is not None:
            self._toMarker.sigDragFinished.disconnect(self._toCallback)

        # clear existing markers
        markers = [
            item.getName()
            for item in plot.getItems()
            if isinstance(item, items.MarkerBase)
        ]
        for marker in markers:
            plot.removeMarker(marker)
        # recreate markers
        plot.addXMarker(
            x=min, legend="from", text="from", color="#3449ebff", draggable=True
        )
        plot.addXMarker(
            x=max, legend="to", text="to", color="#3449ebff", draggable=True
        )
        plot.addXMarker(
            x=estimated_cor,
            legend="estimated cor",
            text="estimated cor",
            color="#34dcebff",
            draggable=True,
        )
        for i_cor, cor in enumerate(other_cors):
            plot.addXMarker(
                x=cor, legend="cor_{}".format(i_cor), draggable=False, color="#34dceb42"
            )

        # handle draggable markers
        self._fromMarker = plot._getMarker("from")
        self._fromCallback = functools.partial(self._limitMarkerMoved, self._fromMarker)
        self._fromMarker.sigDragFinished.connect(self._fromCallback)

        self._toMarker = plot._getMarker("to")
        self._toCallback = functools.partial(self._limitMarkerMoved, self._toMarker)
        self._toMarker.sigDragFinished.connect(self._toCallback)

        self._corMarker = plot._getMarker("estimated cor")
        self._corMarker.sigDragFinished.connect(self._middleMarkerMoved)

    def _middleMarkerMoved(self):
        if self._corMarker is not None:
            self.sigCorEdited.emit(self._corMarker.getXPosition())

    def _limitMarkerMoved(self, marker):
        if self._corMarker is not None and marker is not None:
            new_half_width = numpy.abs(
                marker.getXPosition() - self._corMarker.getXPosition()
            )
            self.sigWidthEdited.emit(new_half_width * 2.0)


class _ReconstructionModeGB(qt.QGroupBox):

    reconstructionModeChanged = qt.Signal(str)

    reconstructionSliceChanged = qt.Signal()

    _DEFAULT_VERTICAL_SLICE_MODE = ("middle", "other")

    def __init__(self, parent=None, title="Slice reconstruction mode"):
        self.__scan = None
        qt.QGroupBox.__init__(self, parent)
        self.setTitle(title)
        self.setLayout(qt.QVBoxLayout())
        self._verticalRB = qt.QRadioButton(ReconstructionMode.VERTICAL.value, self)
        self.layout().addWidget(self._verticalRB)
        self._tiltCorrectionRB = qt.QRadioButton(
            ReconstructionMode.TILT_CORRECTION.value, self
        )
        self.layout().addWidget(self._tiltCorrectionRB)

        # slice choice combobox
        self._defaultSlicesCB = qt.QComboBox(self)
        for mode in self._DEFAULT_VERTICAL_SLICE_MODE:
            self._defaultSlicesCB.addItem(mode)
        self.layout().addWidget(self._defaultSlicesCB)

        # slice selector
        self._sliceVerticalQSB = SliceSelector(self, insert=False, invert_y_axis=True)
        self._sliceVerticalQSB.addSlice(value=0, name="Slice", color="green")
        self._sliceVerticalQSB.setFixedSize(qt.QSize(250, 250))
        self.layout().addWidget(self._sliceVerticalQSB)
        self._sliceTiltCorrQSB = SliceSelector(self, insert=False, invert_y_axis=True)
        self._sliceTiltCorrQSB.addSlice(value=0, name="Slice 0", color="#7e8dff")
        self._sliceTiltCorrQSB.addSlice(value=0, name="Slice 1", color="cyan")
        self._sliceTiltCorrQSB.setFixedSize(qt.QSize(250, 230))
        self.layout().addWidget(self._sliceTiltCorrQSB)
        # select on radio
        # line selector button
        onRadioIcon = silxicons.getQIcon("window-new")
        self._onRadioButton = qt.QPushButton(
            onRadioIcon, "select on radio", parent=self
        )
        self._onRadioButton.setToolTip("Open a dialog to pick the rows from" "a radio")
        self.layout().addWidget(self._onRadioButton)

        # connect signal / slot
        self._verticalRB.toggled.connect(self._reconstructionModeChanged)
        self._tiltCorrectionRB.toggled.connect(self._reconstructionModeChanged)
        self._sliceVerticalQSB.sigSlicesChanged.connect(
            self._reconstructionSliceChanged
        )
        self._onRadioButton.released.connect(self._onRadioActivated)
        self._defaultSlicesCB.currentIndexChanged.connect(self._updateVerticalSelection)

        # set up
        self.blockSignals(True)
        self.setCurrentMode("vertical", "middle")
        self.blockSignals(False)
        # for now we don't have the tilt correction mode
        # self._sliceTiltCorrQSB.hide()
        self._tiltCorrectionRB.hide()

    def setCurrentMode(self, mode, slices):
        valid_modes = ("vertical", "horizontal")
        if mode not in valid_modes:
            raise ValueError(
                "mode {} is not recosgnized. Valid modes are {}"
                "".format(mode, valid_modes)
            )
        self._verticalRB.setChecked(mode == "vertical")
        if mode == "vertical":
            if slices not in self._DEFAULT_VERTICAL_SLICE_MODE:
                raise ValueError(
                    "{} is not a vadid slice mode. Calid are {}"
                    "".format(slices, self._DEFAULT_VERTICAL_SLICE_MODE)
                )
            else:
                index = self._defaultSlicesCB.findText(slices)
                self._defaultSlicesCB.setCurrentIndex(index)
                self._defaultSlicesCB.currentIndexChanged.emit(index)

    def _updateVerticalSelection(self):
        current_mode = self._defaultSlicesCB.currentText()
        mode_is_middle = current_mode == "middle"
        self._sliceVerticalQSB.setVisible(not mode_is_middle)
        self._onRadioButton.setVisible(not mode_is_middle)
        self._reconstructionSliceChanged()

    def _onRadioActivated(self):
        lineSelection = QSliceSelectorDialog(parent=self, n_required_slice=1)
        scan = self.__scan() if self.__scan else None
        if scan is None:
            _logger.info("No scan defined, request user for defining the image" "url")
            dialog = ImageFileDialog()
            if dialog.exec_():
                data = dialog.selectedImage()
            else:
                return
        else:
            assert isinstance(scan, TomwerScanBase)
            try:
                projections = scan.projections
                index_radio = list(projections.keys())[len(projections) // 2]
                data = get_data(projections[index_radio])
            except Exception as e:
                _logger.error("Fail to load radio data:", str(e))
                return

        lineSelection.setData(data)
        lineSelection.setSelection(list(self.getReconstructionSlices().values()))
        if lineSelection.exec_() == qt.QDialog.Accepted:
            selection = lineSelection.getSelection()
            if selection is not None and len(selection) > 0:
                self.setReconstructionSlices({"Slice": selection[0]})

    def setScan(self, scan):
        self.__scan = weakref.ref(scan)

    def _reconstructionSliceChanged(self):
        # for now we only synchronize sliceQSB from vertical with the sinogram
        # viewer
        self.reconstructionSliceChanged.emit()

    def getReconstructionMode(self) -> ReconstructionMode:
        if self._tiltCorrectionRB.isChecked():
            return ReconstructionMode.TILT_CORRECTION
        elif self._verticalRB.isChecked():
            return ReconstructionMode.VERTICAL
        else:
            raise ValueError("No reconstruction mode selected")

    def setReconstructionMode(self, mode: Union[str, ReconstructionMode]) -> None:
        mode = ReconstructionMode.from_value(mode)
        if mode is ReconstructionMode.TILT_CORRECTION:
            self._tiltCorrectionRB.setChecked(True)
        elif mode is ReconstructionMode.VERTICAL:
            self._verticalRB.setChecked(True)
        else:
            raise ValueError("No reconstruction mode selected")

    def getReconstructionSlices(self):
        mode = self.getReconstructionMode()
        if mode is ReconstructionMode.TILT_CORRECTION:
            return self._sliceTiltCorrQSB.getSlicesValue()
        elif mode is ReconstructionMode.VERTICAL:
            if self._defaultSlicesCB.currentText() == "middle":
                return "middle"
            else:
                return self._sliceVerticalQSB.getSlicesValue()
        else:
            raise ValueError("mode {} is not handled".format(mode))

    def setReconstructionSlices(self, slice: Union[dict, int, str]):
        slice_name = None
        if slice == "middle" or isinstance(slice, int) or len(slice) == 1:
            self.setReconstructionMode(ReconstructionMode.VERTICAL)
            target = self._sliceVerticalQSB
            if slice == "middle":
                index = self._defaultSlicesCB.findText("middle")
            else:
                index = self._defaultSlicesCB.findText("other")
                slice_name = "Slice"
            self._defaultSlicesCB.setCurrentIndex(index)
        elif len(slice) == 2:
            self.setReconstructionMode(ReconstructionMode.TILT_CORRECTION)
            target = self._sliceTiltCorrQSB
        else:
            raise ValueError("slice is expected to contain one or two element(s)")
        if slice not in ("middle",):
            if isinstance(slice, int):
                target.setSliceValue(name=slice_name, value=slice)
            else:
                for slice_name, slice_value in slice.items():
                    target.setSliceValue(name=slice_name, value=slice_value)

    def _reconstructionModeChanged(self, *args, **kwargs):
        reconstruction_mode = self.getReconstructionMode()
        self.reconstructionModeChanged.emit(reconstruction_mode.value)
        self._sliceTiltCorrQSB.setVisible(
            reconstruction_mode is ReconstructionMode.TILT_CORRECTION
        )
        self._sliceVerticalQSB.setVisible(
            reconstruction_mode is ReconstructionMode.VERTICAL
        )

    def setSlicesRange(self, min_index, max_index):
        self._sliceTiltCorrQSB.setSlicesRange(min_index, max_index)
        self._sliceVerticalQSB.setSlicesRange(min_index, max_index)

    def getSlicesRange(self):
        return self._sliceVerticalQSB.getSlicesRange()


class _DetectionAccuracyGB(qt.QGroupBox):

    detectionAccuracyChanged = qt.Signal()
    """signal emitted when the value change"""

    _MIN_RANGE = 0.1
    _MAX_RANGE = 4000
    _DEFAULT_VALUE = 10

    def __init__(self, parent=None, title="Detection accuracy (width)"):
        qt.QGroupBox.__init__(self, parent)
        self.setTitle(title)
        self.setLayout(qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.setContentsMargins(0, 0, 0, 0)
        # slider
        self._accuracySlider = LogSlider(parent=self)
        self._accuracySlider.setSuffix(" px")
        self.layout().addWidget(self._accuracySlider, 0, 0, 1, 5)
        # different accuracy buttons
        accuracy_opt_font = self.font()
        accuracy_opt_font.setPixelSize(10)
        self._coarsePB = qt.QPushButton(DetectionAccuracy.COARSE.value, self)
        self._coarsePB.setFont(accuracy_opt_font)
        self._coarsePB.setFlat(True)
        self.layout().addWidget(self._coarsePB, 1, 2, 1, 1)
        self._finePB = qt.QPushButton(DetectionAccuracy.FINE.value, self)
        self._finePB.setFont(accuracy_opt_font)
        self._finePB.setFlat(True)
        self.layout().addWidget(self._finePB, 1, 1, 1, 1)
        self._accuratePB = qt.QPushButton(DetectionAccuracy.ACCURATE.value, self)
        self._accuratePB.setFont(accuracy_opt_font)
        self._accuratePB.setFlat(True)
        self.layout().addWidget(self._accuratePB, 1, 0, 1, 1)

        # connect signal / slot
        self._coarsePB.released.connect(self._setCoarse)
        self._finePB.released.connect(self._setFine)
        self._accuratePB.released.connect(self._setAccurate)
        self._accuracySlider.valueChanged.connect(self.detectionAccuracyChanged)

        # set up
        self._accuracySlider.setRange(self._MIN_RANGE, self._MAX_RANGE)
        self._accuracySlider.setValue(self._DEFAULT_VALUE)

    def _detectionAccuracyChanged(self, *args, **kwargs):
        self.detectionAccuracyChanged.emit()

    def getResearchWidth(self) -> float:
        """Return research width (in pixel)"""
        return self._accuracySlider.value()

    def setResearchWidth(self, width: float):
        """set research width

        :param float width: center of rotation research width
        """
        return self._accuracySlider.setValue(width)

    def _setCoarse(self):
        self._accuracySlider.setValue(ACCURACIES[DetectionAccuracy.COARSE])

    def _setFine(self):
        self._accuracySlider.setValue(ACCURACIES[DetectionAccuracy.FINE])

    def _setAccurate(self):
        self._accuracySlider.setValue(ACCURACIES[DetectionAccuracy.ACCURATE])


class _EstimatedCorWidget(qt.QGroupBox):

    _MIDDLE_COR_TXT = "middle"
    _OTHER_COR_TXT = "other"

    sigAutoCorRequested = qt.Signal()
    """signal emitted when the calculation of automatic cor is asked"""

    sigAutoCorChanged = qt.Signal()
    """signal emitted when the estimated cor value change"""

    def __init__(self, parent=None, title=None):
        assert title is not None
        self._frameWidth = None
        qt.QGroupBox.__init__(self, parent)
        self.setTitle(title)
        self.setLayout(qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self._middlePositionRB = qt.QCheckBox("middle", self)
        self.layout().addWidget(self._middlePositionRB, 0, 0, 1, 2)
        # manual definition of the cor
        self._manualCORAbs = qt.QDoubleSpinBox(self)
        self._manualCORAbs.setDecimals(4)
        self._manualCORAbs.setMaximum(9999999)
        self._manualCORAbs.setSuffix(" px")
        self._absLabel = qt.QLabel("absolute: ")
        self.layout().addWidget(self._absLabel, 1, 0, 1, 1)
        self.layout().addWidget(self._manualCORAbs, 1, 1, 1, 1)
        self._manualCORRel = qt.QDoubleSpinBox(self)
        self._manualCORRel.setDecimals(4)
        self._manualCORRel.setSuffix(" px")
        self._manualCORRel.setMinimum(-9999999)
        self._manualCORRel.setMaximum(9999999)
        self._relLabel = qt.QLabel("relative: ")
        self.layout().addWidget(self._relLabel, 2, 0, 1, 1)
        self.layout().addWidget(self._manualCORRel, 2, 1, 1, 1)
        # auto cor button
        self._autoCorPB = qt.QPushButton("auto cor", self)
        auto_cor_icon = icons.getQIcon("a")
        self._autoCorPB.setIcon(auto_cor_icon)
        self.layout().addWidget(self._autoCorPB, 3, 1, 1, 1)

        # connect signal / slot
        self._middlePositionRB.toggled.connect(self._modeChanged)
        self._manualCORRel.editingFinished.connect(self._relativePositionChanged)
        self._manualCORAbs.editingFinished.connect(self._absolutePositionChanged)
        self._autoCorPB.released.connect(self._autoCorReqested)
        self._middlePositionRB.toggled.connect(self._autoCorValueChanged)
        self._manualCORRel.editingFinished.connect(self._autoCorValueChanged)
        self._manualCORAbs.editingFinished.connect(self._autoCorValueChanged)
        # set up
        self._middlePositionRB.setChecked(True)
        self._modeChanged()

    def getCorValue(self, mode="relative"):
        assert mode in ("rel", "relative", "abs", "absolute")
        if self._middlePositionRB.isChecked():
            if mode in ("rel", "relative"):
                return 0.0
            elif mode in ("abs", "absolute"):
                if self._frameWidth is not None:
                    return self._frameWidth / 2.0
                else:
                    return self._MIDDLE_COR_TXT
            else:
                raise ValueError("Mode {} is not recognized".format(mode))
        else:
            if mode in ("rel", "relative"):
                return self._manualCORRel.value()
            elif mode in ("abs", "absolute"):
                return self._manualCORAbs.value()
            else:
                raise ValueError("Mode {} is not recognized".format(mode))

    def _modeChanged(self, *args, **kwargs):
        manual_opt_visible = not self._middlePositionRB.isChecked()
        self._absLabel.setVisible(manual_opt_visible)
        self._relLabel.setVisible(manual_opt_visible)
        self._manualCORRel.setVisible(manual_opt_visible)
        self._manualCORAbs.setVisible(manual_opt_visible)

    def _relativePositionChanged(self, *args, **kwargs):
        if self._frameWidth is not None:
            old = self.blockSignals(True)
            old_mcor = self._manualCORAbs.blockSignals(True)
            self._manualCORAbs.setValue(
                self._manualCORRel.value() + self._frameWidth / 2.0
            )
            self._manualCORAbs.blockSignals(old_mcor)
            self.blockSignals(old)

    def _absolutePositionChanged(self, *args, **kwargs):
        if self._frameWidth is not None:
            old = self.blockSignals(True)
            old_mcor = self._manualCORRel.blockSignals(True)
            self._manualCORRel.setValue(
                self._manualCORAbs.value() - self._frameWidth / 2.0
            )
            self._manualCORRel.blockSignals(old_mcor)
            self.blockSignals(old)

    def setFrameWidth(self, width):
        self._frameWidth = width
        # if necessary auto reset of the absolute value
        if self._manualCORAbs.value() == self._manualCORRel.value():
            self._relativePositionChanged()

    def getFrameWidth(self):
        return self._frameWidth

    def setCorValue(self, value):
        """

        :param Union[str,float] value: if a float expected to be given as the
                                       relative value
        :return:
        """
        if value == self._MIDDLE_COR_TXT:
            self._middlePositionRB.setChecked(True)
        else:
            self._middlePositionRB.setChecked(False)
            self._manualCORRel.setValue(float(value))
            self._manualCORRel.editingFinished.emit()

    def setAbsoluteCorValue(self, value):
        if value == self._MIDDLE_COR_TXT:
            self._middlePositionRB.setChecked(True)
        else:
            self._middlePositionRB.setChecked(False)
            self._manualCORAbs.setValue(float(value))
            self._manualCORAbs.editingFinished.emit()

    def _autoCorReqested(self):
        self.sigAutoCorRequested.emit()

    def _autoCorValueChanged(self):
        self.sigAutoCorChanged.emit()


class SAAxisOptions(qt.QWidget):
    """
    Options of the semi-automatic center of rotation detection
    """

    sigReconstructionSliceChanged = qt.Signal()
    """signal emitted when the slice to reconstruct change"""

    sigReconstructionRangeChanged = qt.Signal()
    """signal emitted when the reconstruction parameters change"""

    sigConfigurationChanged = qt.Signal()
    """signal emitted when the configuration change"""

    _FONT_IMPORTANT = qt.QFont("Arial", 12)
    _FONT_MEDIUM = qt.QFont("Arial", 10)

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        # reconstruction mode
        self._reconstructionMode = _ReconstructionModeGB(parent=self)
        self._reconstructionMode.setFont(self._FONT_MEDIUM)
        self.layout().addWidget(self._reconstructionMode)

        # estimated cor
        self._estimatedCorWidget = _EstimatedCorWidget(
            self,
            title="Estimated cor position (x axis)",
        )
        self._estimatedCorWidget.setFont(self._FONT_IMPORTANT)
        self.layout().addWidget(self._estimatedCorWidget)

        # detection accuracy
        self._detectionAccuracy = _DetectionAccuracyGB(parent=self)
        self._detectionAccuracy.setFont(self._FONT_MEDIUM)
        self.layout().addWidget(self._detectionAccuracy)

        # number of reconstructions eq volume size
        self._nReconsWidget = qt.QWidget(parent)
        self._nReconsWidget.setLayout(qt.QFormLayout())
        self._nReconsWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._nReconsSB = qt.QSpinBox(self._nReconsWidget)
        self._nReconsSB.setRange(0, 2000)
        self._nReconsSB.setValue(30)
        self._nReconsWidget.setFont(self._FONT_IMPORTANT)
        self._nReconsWidget.layout().addRow("Number of reconstruction", self._nReconsSB)
        self.layout().addWidget(self._nReconsWidget)

        # spacer
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        # expose API
        self.sigAutoCorRequested = self._estimatedCorWidget.sigAutoCorRequested

        # connect signal / slot
        self._reconstructionMode.reconstructionSliceChanged.connect(
            self._reconstructionSliceChanged
        )
        self._nReconsSB.valueChanged.connect(self._reconstructionRangeChanged)
        self._estimatedCorWidget.sigAutoCorChanged.connect(
            self._reconstructionRangeChanged
        )
        self._detectionAccuracy.detectionAccuracyChanged.connect(
            self._reconstructionRangeChanged
        )

    def setDetectionAccuracy(self, width):
        if width != self._detectionAccuracy.getResearchWidth():
            self._detectionAccuracy.setResearchWidth(width)
            self.sigConfigurationChanged.emit()

    def setEstimatedCorPosition(self, value):
        if value != self._estimatedCorWidget.getCorValue("relative"):
            self._estimatedCorWidget.setCorValue(value)
            self.sigConfigurationChanged.emit()

    def setEstimatedAbsoluteCorPosition(self, value):
        if value != self._estimatedCorWidget.getCorValue("absolute"):
            self._estimatedCorWidget.setAbsoluteCorValue(value)
            self.sigConfigurationChanged.emit()

    def getEstimatedCorPosition(self, *args, **kwargs):
        return self._estimatedCorWidget.getCorValue(*args, **kwargs)

    def setScan(self, scan: TomwerScanBase):
        if scan is not None:
            self.setSlicesRange(0, scan.dim_2)
            self._estimatedCorWidget.setFrameWidth(scan.dim_1)
            self._reconstructionMode.setScan(scan)

    def getFrameWidth(self):
        return self._estimatedCorWidget.getFrameWidth()

    def setFrameWidth(self, width):
        if width != self._estimatedCorWidget.getFrameWidth():
            self._estimatedCorWidget.setFrameWidth(width)
            self.sigConfigurationChanged.emit()

    def getSlicesRange(self) -> tuple:
        self._reconstructionMode.getSlicesRange()

    def setSlicesRange(self, min_index, max_index):
        self._reconstructionMode.setSlicesRange(min_index, max_index)
        self.sigConfigurationChanged.emit()

    def _reconstructionSliceChanged(self):
        self.sigReconstructionSliceChanged.emit()

    def _reconstructionRangeChanged(self):
        self.sigReconstructionRangeChanged.emit()
        self.sigConfigurationChanged.emit()

    def getReconstructionSlices(self) -> tuple:
        return self._reconstructionMode.getReconstructionSlices()

    def setReconstructionSlices(self, slices) -> tuple:
        return self._reconstructionMode.setReconstructionSlices(slices)

    def getReconstructionMode(self):
        return self._reconstructionMode.getReconstructionMode()

    def setReconstructionMode(self, mode):
        self._reconstructionMode.setReconstructionMode(mode)
        self.sigConfigurationChanged.emit()

    def getNReconstruction(self):
        return self._nReconsSB.value()

    def setNReconstruction(self, n):
        self._nReconsSB.setValue(int(n))
        self.sigConfigurationChanged.emit()

    def getResearchWidth(self):
        return self._detectionAccuracy.getResearchWidth()

    def setResearchWidth(self, width):
        self._detectionAccuracy.setResearchWidth(width)
        self.sigConfigurationChanged.emit()


class SliceAndCorWidget(qt.QMainWindow):
    sigConfigurationChanged = qt.Signal()
    """signal emitted when the configuration changed"""

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        # sinogram viewer
        self._sinogramViewer = SinogramViewer(self)
        self._sinogramViewer.setWindowFlags(qt.Qt.Widget)
        self.setCentralWidget(self._sinogramViewer)
        # slice and accuracy options
        self._saaxisOpts = SAAxisOptions(self)
        self._dockWidgetOpts = qt.QDockWidget(self)
        self._dockWidgetOpts.setFixedWidth(300)
        self._dockWidgetOpts.setWidget(self._saaxisOpts)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._dockWidgetOpts)
        self._dockWidgetOpts.setFeatures(qt.QDockWidget.DockWidgetMovable)
        # sinogram options
        self.sinoOprsDW = self._sinogramViewer.getOptionsDockWidget()
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self.sinoOprsDW)
        self.sinoOprsDW.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.sinoOprsDW.setFixedWidth(300)

        # expose API
        self.getEstimatedCorPosition = self._saaxisOpts.getEstimatedCorPosition
        self.setEstimatedCorPosition = self._saaxisOpts.setEstimatedCorPosition
        self.setEstimatedAbsoluteCorPosition = (
            self._saaxisOpts.setEstimatedAbsoluteCorPosition
        )
        self.setDetectionAccuracy = self._saaxisOpts.setDetectionAccuracy
        self.getNReconstruction = self._saaxisOpts.getNReconstruction
        self.setNReconstruction = self._saaxisOpts.setNReconstruction
        self.getResearchWidth = self._saaxisOpts.getResearchWidth
        self.setResearchWidth = self._saaxisOpts.setResearchWidth
        self.getReconstructionSlices = self._saaxisOpts.getReconstructionSlices
        self.setReconstructionSlices = self._saaxisOpts.setReconstructionSlices
        self.getReconstructionMode = self._saaxisOpts.getReconstructionMode
        self.setReconstructionMode = self._saaxisOpts.setReconstructionMode
        self.getFrameWidth = self._saaxisOpts.getFrameWidth
        self.setFrameWidth = self._saaxisOpts.setFrameWidth
        self.getSlicesRange = self._saaxisOpts.getSlicesRange
        self.setSlicesRange = self._saaxisOpts.setSlicesRange
        self.sigStartSinogramLoad = self._sinogramViewer.sigSinoLoadStarted
        self.sigEndSinogramLoad = self._sinogramViewer.sigSinoLoadEnded
        self.sigConfigurationChanged = self._saaxisOpts.sigConfigurationChanged

        # expose signals
        self.sigReconstructionSliceChanged = (
            self._saaxisOpts.sigReconstructionSliceChanged
        )
        self.sigAutoCorRequested = self._saaxisOpts.sigAutoCorRequested
        self.sigReconstructionRangeChanged = (
            self._saaxisOpts.sigReconstructionRangeChanged
        )

        # connect signal / slot
        self._sinogramViewer.sigCorEdited.connect(self.setEstimatedAbsoluteCorPosition)
        self._sinogramViewer.sigWidthEdited.connect(self.setDetectionAccuracy)

    def loadSinogram(self):
        """Enforce to load a sinogram with the current settings"""
        self._sinogramViewer.loadSinogram()

    def setScan(self, scan):
        self._sinogramViewer.setScan(scan, update=False)
        self._saaxisOpts.setScan(scan)

    def close(self):
        self._sinogramViewer.close()
        self._sinogramViewer = None
        super().close()


if __name__ == "__main__":
    import tempfile
    import shutil
    from tomwer.core.utils.scanutils import MockHDF5

    app = qt.QApplication([])
    window = SliceAndCorWidget()

    img_width = 216
    slices_folder = tempfile.mkdtemp()

    scan = MockHDF5(
        scan_path=slices_folder,
        n_proj=60,
        n_ini_proj=60,
        create_ini_dark=False,
        create_ini_ref=False,
        dim=img_width,
    ).scan

    window.setScan(scan)
    window.setWindowTitle("cor and slice selection")
    window.show()
    app.exec_()

    shutil.rmtree(slices_folder)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "01/12/2016"

from silx.gui import qt
from orangewidget import widget, gui
from orangewidget.widget import Input
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.gui.viewerqwidget import ScanWidget
from tomwer.core.scan.scanbase import TomwerScanBase
import tomwer.core.process.visualization.imagestackviewer
import logging

logger = logging.getLogger(__name__)


class ImageStackViewerOW(widget.OWBaseWidget, openclass=True):
    """a simple viewer of image stack

    :param parent:the parent widget
    :param FtserieReconstruction ftseries: the initial reconstruction to show
    """

    name = "(old) data viewer"
    id = "orange.widgets.tomwer.imagestackviewer"
    description = "Show a small recap of the reconstruction runned"
    icon = "icons/eyecrack.png"
    priority = 700
    keywords = ["tomography", "file", "tomwer", "acquisition", "validation"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    ewokstask = (
        tomwer.core.process.visualization.imagestackviewer._ImageStackViewerPlaceHolder
    )

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)

    def __init__(self, parent=None, ftseries=None):
        widget.OWBaseWidget.__init__(self, parent)

        self.tabsWidget = {}
        # building GUI
        self._scanWidgetLayout = gui.vBox(self.mainArea, self.name).layout()
        self.viewer = ScanWidget(parent=self, canLoadOtherScan=False)

        self._scanWidgetLayout.addWidget(self.viewer)
        self._scanWidgetLayout.setContentsMargins(0, 0, 0, 0)
        if ftseries is not None:
            self.viewer.updateData(ftseries)

    @Inputs.data
    def addScan(self, ftseriereconstruction):
        if ftseriereconstruction is None:
            return

        _ftserie = ftseriereconstruction
        if type(ftseriereconstruction) is str:
            _ftserie = ScanFactory.create_scan_object(_ftserie)
        assert isinstance(_ftserie, TomwerScanBase)
        return self.viewer.updateData(_ftserie)

    def updateFromPath(self, path):
        if path is not None:
            return self.viewer.updateFromPath(path)

    def sizeHint(self):
        return qt.QSize(400, 500)
